;**********************************************************
;***  	Universal Light Controller Version 2.0       
;***   	This file contains:
;***	Processor spec. definitions, port usage, FSRs etc.
;**********************************************************

;Port usage (12F1822):
;RA3 = input: RC impulse, channel 1
;RA5 = input: RC impulse, ch.2 (if RCCHAN =2)
;RA0..RA2, RA4, RA5: LED outputs (max. 5 channels)
	


;================================================================================
;Definitions for PIC12F1822:
  ifdef __12F1822
	#define PINS_8
	#include <P12F1822.INC>         	; processor specific variable definitions
    ifdef SET_CONFIG
	__config _CONFIG1, _FOSC_INTOSC & _WDTE_OFF & _PWRTE_ON & _MCLRE_OFF & _CP_OFF & _BOREN_ON & _IESO_OFF & 0x3FFF
	__config _CONFIG2, _WRT_OFF & _PLLEN_ON & _BORV_HI & _LVP_OFF & 0x3FFF
    endif
	if (RCCHAN > 1)
	  #define MAX_NCHAN 4
	else
	  #define MAX_NCHAN 5
	endif

	#define CLOCK 32
  endif

;================================================================================
;Definitions for PIC12F1840:
  ifdef __12F1840
	#define PINS_8
	#include <P12F1840.INC>         	; processor specific variable definitions
    ifdef SET_CONFIG
	__config _CONFIG1, _FOSC_INTOSC & _WDTE_OFF & _PWRTE_ON & _MCLRE_OFF & _CP_OFF & _BOREN_ON & _IESO_OFF & 0x3FFF
	__config _CONFIG2, _WRT_OFF & _PLLEN_ON & _BORV_HI & _LVP_OFF & 0x3FFF
    endif
	if (RCCHAN > 1)
	  #define MAX_NCHAN 4
	else
	  #define MAX_NCHAN 5
	endif

	#define CLOCK 32
  endif

;================================================================================
;Definitions for PIC16F1823:
  ifdef __16F1823
	#define PINS_14
	#include <P16F1823.INC>         	; processor specific variable definitions
    ifdef SET_CONFIG
	__config _CONFIG1, _FOSC_INTOSC & _WDTE_OFF & _PWRTE_ON & _MCLRE_OFF & _CP_OFF & _BOREN_ON & _IESO_OFF & 0x3FFF
	__config _CONFIG2, _WRT_OFF & _PLLEN_ON & _BORV_HI & _LVP_OFF & 0x3FFF
    endif
	#define MAX_NCHAN 8
	#define CLOCK 32
  endif

;================================================================================
;Definitions for PIC16F1825:
  ifdef __16F1825
	#define PINS_14
	#include <P16F1825.INC>         	; processor specific variable definitions
    ifdef SET_CONFIG
	__config _CONFIG1, _FOSC_INTOSC & _WDTE_OFF & _PWRTE_ON & _MCLRE_OFF & _CP_OFF & _BOREN_ON & _IESO_OFF & 0x3FFF
	__config _CONFIG2, _WRT_OFF & _PLLEN_ON & _BORV_HI & _LVP_OFF & 0x3FFF
    endif
	#define MAX_NCHAN 8
	#define CLOCK 32
  endif

;================================================================================
;Definitions for PIC16F1826:
  ifdef __16F1826
	#define PINS_18
	#include <P16F1826.INC>         	; processor specific variable definitions
    ifdef SET_CONFIG
	__config _CONFIG1, _FOSC_INTOSC & _WDTE_OFF & _PWRTE_ON & _MCLRE_OFF & _CP_OFF & _BOREN_ON & _IESO_OFF & 0x3FFF
	__config _CONFIG2, _WRT_OFF & _PLLEN_ON & _BORV_HI & _LVP_OFF & 0x3FFF
    endif
	#define MAX_NCHAN 8
	#define CLOCK 32
  endif

;================================================================================

 if NCHAN > MAX_NCHAN
	error "Number of output channels out of range!"
 endif 

;=================================================================================
;Pinout definitions for several packages:

  ifdef PINS_8
	#define	PORT_CH0 PORTA
	#define PIN_CH0 0
	#define	PORT_CH1 PORTA
	#define PIN_CH1 1
	#define	PORT_CH2 PORTA
	#define PIN_CH2 2
	#define	PORT_CH3 PORTA
	#define PIN_CH3 4
	#define	PORT_CH4 PORTA
	#define PIN_CH4 5

	#define PORT_RX0 PORTA
	#define PIN_RX0 3
	#define PORT_RX1 PORTA
	#define PIN_RX1 5
	#define IOCP IOCAP
	#define IOCN IOCAN
	#define IOCFLG IOCAF
  endif

  ifdef PINS_14
	#define	PORT_CH0 PORTC
	#define PIN_CH0 0
	#define	PORT_CH1 PORTC
	#define PIN_CH1 1
	#define	PORT_CH2 PORTC
	#define PIN_CH2 2
	#define	PORT_CH3 PORTC
	#define PIN_CH3 3
	#define	PORT_CH4 PORTC
	#define PIN_CH4 4
	#define	PORT_CH5 PORTC
	#define PIN_CH5 5
	#define	PORT_CH6 PORTA
	#define PIN_CH6 4
	#define	PORT_CH7 PORTA
	#define PIN_CH7 5

	#define PORT_RX0 PORTA
	#define PIN_RX0 3
	#define PORT_RX1 PORTA
	#define PIN_RX1 0
	#define IOCP IOCAP
	#define IOCN IOCAN
	#define IOCFLG IOCAF

  endif

  ifdef PINS_18
	#define	PORT_CH0 PORTB
	#define PIN_CH0 4
	#define	PORT_CH1 PORTB
	#define PIN_CH1 5
	#define	PORT_CH2 PORTB
	#define PIN_CH2 6
	#define	PORT_CH3 PORTB
	#define PIN_CH3 7
	#define	PORT_CH4 PORTA
	#define PIN_CH4 6
	#define	PORT_CH5 PORTA
	#define PIN_CH5 7
	#define	PORT_CH6 PORTA
	#define PIN_CH6 0
	#define	PORT_CH7 PORTA
	#define PIN_CH7 1

	#define PORT_RX0 PORTB
	#define PIN_RX0 0
	#define IOCP IOCBP
	#define IOCN IOCBN
	#define IOCFLG IOCBF

  endif