

;*********************************************************************
;**  ASSEMBLER MACRO DEFINITIONS OF ALIC INSTRUCTIONS
;*********************************************************************

;Macro: checks if (addr) = EEPROM address:
chkaddr	macro addr
 variable a=addr-entry_ch0	; causes error if addr is no valid jump-label
	endm

;Used to create variable adressing (special parameter range 0xF0..0xFF)
#define VAR(x) (x + 0xA5F0)

;Macro: checks valid parameter range 
;sets variable "param" according to given value
;user may either specify a constant value in range 0..239
;or specify a variable by expression "VAR(x)", which will finally map to
;240..255 (for x = 0..15)
chkparameter macro p
  if p < 240
param	set p		;range 0..239: param is constant value
  else
param 	set 0xFFFF	;if given variable is out of valid range... 
    if p < 0xA5F0
	exitm		;invalid range!
    endif
    if p > 0xA5FF    
	exitm		;invalid range!
    endif
param	set (p & 0xFF)	;valid range for var: param = 0xF0..0xFF
  endif
	endm	
 
;ALIC Opcode macro definitions:

ADDCV	macro	value
	de	OPC_ADDCV,value
	endm

ADDCVX	macro	value,destvar
	de	OPC_ADDCVX,value,destvar
	endm

ADDV	macro	src
	de	OPC_ADDV,src
	endm

ADDVX	macro	src,dest
	de	OPC_ADDVX,src,dest
	endm

#define SETBRIGHT BRIGHT	;(for compatibility)
BRIGHT macro value,delay
 if delay>255
	error "BRIGHT: Delay out of range"
 endif
 if value>=0xC0
	error "BRIGHT: Brightness value out of range"
 endif
	de	value,delay
	endm

BRIGHTV macro	ticks
	de	OPC_BRIGHTV,ticks
	endm

BRIGHTVX	macro 	var,ticks
	de	OPC_BRIGHTVX,var,ticks
	endm

CLIPV	macro	min,max
	local p1
	chkparameter min
  if param > 255
	error "CLIPV: Parameter (min) value or variable out of range!"
  endif
p1	set	param
	chkparameter max
  if param > 255
	error "CLIPV: Parameter (max) value or variable out of range!"
  endif	
	de	OPC_CLIPV,p1,param
	endm

CLIPVX	macro	min,max,var
	local p1
	chkparameter min
  if param > 255
	error "CLIPVX: Parameter (min) value or variable out of range!"
  endif
p1	set	param
	chkparameter max
  if param > 255
	error "CLIPVX: Parameter (max) value or variable out of range!"
  endif	
	de	OPC_CLIPVX,p1,param,var
	endm

#define CLRSW CLRS		;(for compatibility)
CLRS 	macro
	de	OPC_CLRS
	endm

#define CLRSWX CLRSX		;(for compatibility)
CLRSX	macro	switch
	de	OPC_CLRSX,switch
	endm

COPYV	macro	sourcevar
	de	OPC_COPYV,sourcevar
	endm

COPYVX	macro	sourcevar,destvar
	de	OPC_COPYVX,sourcevar,destvar
	endm

DEFBRIGHT macro value
	chkparameter ms
  if param > 255
	error "DEFBRIGHT: Parameter value or variable out of range!"
  endif

	de	OPC_DEFBRIGHT,param
	endm

DEFSWITCH macro switch
	de	OPC_DEFSWITCH,switch
	endm

DEFTAU macro ms
  if ms>16000
	error "DEFTAU: parameter exceeds maximum time (16s)"
  endif	
	dw	OPC_DEFTAU,ms
	endm

DEFTICKLEN macro ms
	chkparameter ms
  if param > 255
	error "DEFTICKLEN: Parameter value or variable out of range!"
  endif
	de	OPC_DEFTICKLEN,param
	endm

DEFVAR	macro	var
	de	OPC_DEFVAR,var
	endm

DELAY 	macro	ticks
	chkparameter ticks
  if param > 255
	error "DELAY: Parameter value or variable out of range!"
  endif
	de	OPC_DELAY,param
	endm

#define HALTSWOFF HALTSOFF	;(for compatibility)
HALTSOFF macro
	de	OPC_HALTSOFF
	endm

HALTSXOFF macro switch
	de	OPC_HALTSXOFF,switch
	endm

#define HALTSWON HALTSON	;(for compatibility)
HALTSON macro
	de	OPC_HALTSON
	endm

HALTSXON macro	switch
	de	OPC_HALTSXON,switch
	endm

#define JSWOFF JSOFF		;(for compatibility)
JSOFF 	macro	addr
	chkaddr addr
	dw	OPC_JSOFF
	dw 	addr
	endm

#define JSWON JSON		;(for compatibility)
JSON 	macro	addr
	chkaddr addr
	dw	OPC_JSON,addr
	endm

#define JSWXOFF JSXOFF		;(for compatibility)
JSXOFF	macro	switch,addr
	chkaddr	addr
	dw	OPC_JSXOFF,switch,addr
	endm

#define JSWXON JSXON		;(for compatibility)
JSXON	macro	switch,addr
	chkaddr	addr
	dw	OPC_JSXON,switch,addr
	endm

JUMP 	macro	addr
	chkaddr	addr
	dw	OPC_JUMP,addr
	endm

#define JVARAE JVAE		;(for compatibility)
JVAE	macro	val,addr
	chkaddr addr
	dw	OPC_JVAE,val,addr
	endm

#define JVARB JVB		;(for compatibility)
JVB	macro	val,addr
	chkaddr addr
	dw	OPC_JVB,val,addr
	endm

#define JVARXAE JVXAE		;(for compatibility)
JVXAE macro 	var,value,addr
	chkaddr	addr
	dw 	OPC_JVXAE,var,value,addr
	endm

#define JVARXB JVXB		;(for compatibility)
JVXB	macro	var,value,addr
	chkaddr	addr
	dw	OPC_JVXB,var,value,addr
	endm

LIGHTOFF macro delay
 if delay>255
	error "LIGHTOFF: Delay out of range"
 endif
	de	OPC_LIGHTOFF,delay
	endm

LIGHTON macro delay
 if delay>255
	error "LIGHTON: Delay out of range"
 endif
	de	OPC_LIGHTON,delay
	endm

MULV	macro	src
	de	OPC_MULV,src
	endm

MULVX	macro	src,dest
	de	OPC_MULVX,src,dest
	endm

RNDBRT	macro 	min,max,delay
	local p1
	chkparameter min
  if param > 255
	error "RNDBRT: Parameter (min) value or variable out of range!"
  endif
p1	set	param
	chkparameter max
  if param > 255
	error "RNDBRT: Parameter (max) value or variable out of range!"
  endif
; if (min >= 0xC0) || (max>=0xC0)
;	error "RNDBRT: Brightness out of range"
; endif
	de	OPC_RNDBRT,p1,param,delay
	endm

RNDDLY 	macro	maxticks
	de	OPC_RNDDLY,maxticks
	endm

RNDJMP	macro	chance,addr
	chkparameter chance
  if param > 255
	error "RNDJMP: Parameter value or variable out of range!"
  endif
	chkaddr	addr
	dw	OPC_RNDJMP,param,addr
	endm

RNDV	macro	min,max
	local p1
	chkparameter min
  if param > 255
	error "RNDV: Parameter (min) value or variable out of range!"
  endif
p1	set	param
	chkparameter max
  if param > 255
	error "RNDV: Parameter (max) value or variable out of range!"
  endif
	de	OPC_RNDV,p1,param
	endm

RNDVX	macro	min,max,dest
	local p1
	chkparameter min
  if param > 255
	error "RNDVX: Parameter (min) value or variable out of range!"
  endif
p1	set	param
	chkparameter max
  if param > 255
	error "RNDVX: Parameter (max) value or variable out of range!"
  endif
	de	OPC_RNDVX,p1,param,dest
	endm

SCLDWNV macro	min,max
	local p1
	chkparameter min
  if param > 255
	error "SCLDWNV: Parameter (min) value or variable out of range!"
  endif
p1	set	param
	chkparameter max
  if param > 255
	error "SCLDWNV: Parameter (max) value or variable out of range!"
  endif
	de	OPC_SCLDWNV,p1,param
	endm

SCLDWNVX macro min,max,var
	local p1
	chkparameter min
  if param > 255
	error "SCLDWNVX: Parameter (min) value or variable out of range!"
  endif
p1	set	param
	chkparameter max
  if param > 255
	error "SCLDWNVX: Parameter (max) value or variable out of range!"
  endif
	de	OPC_SCLDWNVX,p1,param,var
	endm

SCLUPV 	macro	min,max
	local p1
	chkparameter min
  if param > 255
	error "SCLUPV: Parameter (min) value or variable out of range!"
  endif
p1	set	param
	chkparameter max
  if param > 255
	error "SCLUPV: Parameter (max) value or variable out of range!"
  endif
	de	OPC_SCLUPV,p1,param
	endm

SCLUPVX macro min,max,var
	local p1
	chkparameter min
  if param > 255
	error "SCLUPVX: Parameter (min) value or variable out of range!"
  endif
p1	set	param
	chkparameter max
  if param > 255
	error "SCLUPVX: Parameter (max) value or variable out of range!"
  endif
	de	OPC_SCLUPVX,p1,param,var
	endm

#define SETSW SETS		;(for compatibility)
SETS 	macro
	de	OPC_SETS
	endm

#define SETSWX SETSX		;(for compatibility)
SETSX	macro	switch
	de	OPC_SETSX,switch
	endm

SETV 	macro	value
	de	OPC_SETV,value
	endm

SETVX 	macro	value,var
	de	OPC_SETVX,value,var
	endm

SMOOTHOFF macro
	de	OPC_SMOOTHOFF
	endm

SMOOTHON macro
	de	OPC_SMOOTHON
	endm

P_RESET macro
	de	OPC_P_RESET
	endm

;**************************************************************