;**********************************************************
;***  	Universal Light Controller EM			***
;***	Version 0.4       				***
;**********************************************************
;supported PICs:

;
;History:
;0.0 initial Beta-Version
;0.1 Smooth-Mode Bug fixed, added "#define PWM_DITHER" option

;Timer usage: 
;	TMR0: PWM generation
;	TMR1: RC pulse processing
;	TMR2: system timer

#include picalic_config.inc
#define SET_CONFIG
#include macros.inc
#include MCU_Headers_em.inc	;Pocessor specific definitions
#include alic_opcodes.inc	;Opcodes and macros for ALIC instructions
#include alic_macros.inc

;special bits:
#define ADRMODE 0		;opcode.addressing mode: 0= dedicated/1= address from parameters

;____________________________________________________________
;		  			Bank 0 RAM Data								 
;____________________________________________________________

TEMP0 	udata_ovr 0x20		;-------shared memory: AdjustPW destroys content! -----
opcode 	res 1			;stores current opcode
y1_low 	res 1			;used for arithmetic (interpolation, multiplication...) 
y1_high res 1
y2_low 	res 1
y2_high res 1
x1 	res 1
x2 	res 1
x_low	res 1			
x_high 	res 1
saveTCnt res 1
save_ip res 2			

DATA0	udata 	0x2C

tickcnt res 1			;increments every millisecond, decremeted by tick-processing
ip_low	res 1			;instruction pointer of currently selected channel processor
ip_high	res 1			
channel	res 1			;currently processed output channel (0..NCHAN-1)
y_low	res 1
y_high 	res 1

switch res 1			;8 control switches (1 per bit)
random res 1			;8 bit random number
varbase res NVARS		;reserve space for control variables (except RC-vars)

#define x x_low
#define y y_low

;_____________________________________________________________
;				RAM accessable in all banks:
;_____________________________________________________________

	udata_shr
temp	res 3		;temporary data
w16	res 2		;16 bit working register

;_____________________________________________________________
;			Channel processing variables structure:
;_____________________________________________________________
	CBLOCK 0	;(labeled constants are the offset!)
c_Flags			;flag bits
c_IPL			;instruction pointer low
c_IPH			;instruction pointer high
c_msCnt			;milliseconds downcounter (->ticklength)
c_TCnt			;tick counter for delays (in ticks)
;for interpolation (smooth):
c_Period		;full length (ticks) of current period
c_val1L			;pwm @beginning of period
c_val1H
c_val2L			;pwm @end of period
c_val2H	
c_tau			;time constant for exponential decay
			;bits 0..3: step time 	16..31 ms (if b7=0)
			;			64..124 ms (if b7=1)
			;bits 4..6: number of steps/T (power of 2)
c_tauCnt		;millisecond counter
c_pwdL			;pwm drive value	
c_pwdH			;(exponential decay will approach this value)
c_pwoL			;current output pulse width
c_pwoH			
	
PV_SIZE:0
	ENDC

;c_Flags bit numbers:
#define SMOOTH_EN 0	;enable smooth mode	
#define SMOOTHEOP 1	;smooth scanning: found end of period
#define SMOOTHFLG 2	;smooth mode currently active
#define JMPFLG	3	;last instruction was wait or jump (short delay!)

LINEAR0	udata
ChProcVars res NCHAN*PV_SIZE	;space for NCHAN * Processing Variables

	global temp,w16,channel,tickcnt
	extern AdjustPW, initPWM
 if RCCHAN
	extern rc_init, procRCIn, memswtch, rcdetect, rcval
 endif


;______________________________________________________________
;**************************************************************
;		  			Program Memory starts here:					
;**************************************************************

;========== Reset Entry:
RESETV	code	0
start	
	pagesel init
 	goto	init

;____________________________________________________________________
; 	 		  			Device Initialization						
;____________________________________________________________________
PROG0 	code			
 errorlevel-302

init
 if ACTIVELOW
	movlw	0xFF		;active low outputs: 1 = off
 else
	movlw	0x00		;active high outputs: 0 = off
 endif
	movwf	PORTA		;all LEDs off

 #ifdef PORTB
	movwf	PORTB		;all LEDs off
 #endif
 #ifdef PORTB
	movwf	PORTC		;all LEDs off
 #endif

;after reset, MPU is running at 500 kHz clock
;set clock to 32MHz:
	banksel	OSCCON
	movlw	b'01110000'
	movwf	OSCCON

	banksel	ANSELA
	clrf	ANSELA		;all Pins = digital I/O
 #ifdef ANSELB
	movwf	ANSELB		;all Pins = digital I/O
 #endif
 #ifdef ANSELC
	movwf	ANSELC		;all Pins = digital I/O
 #endif

;--------------- set TRIS register(s)
	banksel TRISA			
;***********************************************************
  ifdef PINS_8			;Pinout = 8 pins package: 
trisval	set	~AUXOUTA
     if NCHAN <4
trisval &= (0xFF << NCHAN)	;RA0/RA1/RA2 = first 3 outputs
     else
trisval &= (0xFF << NCHAN+1)	;skip RA3 (=always input)
     endif
     if RCCHAN == 2
trisval |=	B'100000'	;RA5 = input (RC #2)
     endif
	movlw	low trisval
	movwf	TRISA
  endif					
;***********************************************************
  ifdef PINS_14			;Pinout = 14 pins package: 
trisval	set	~AUXOUTA
     if RCCHAN == 2
trisval |=	1		;RA0 = input (RC #2)
     endif
     if NCHAN > 6
trisval &=	B'101111'	;RA4 = output ch6
     endif
     if NCHAN > 7	
trisval &=	B'011111'	;RA5 = output ch7
     endif
	movlw	trisval
	movwf	TRISA

trisval	set	~AUXOUTC
trisval &= (0xFF << NCHAN)
	movlw	low trisval
	movwf	TRISC
  endif						
;***********************************************************
  ifdef PINS_18			;Pinout = 18 pins package: 
trisval	set	~AUXOUTA
   if NCHAN >4
trisval &=	B'10111111'	;RA6 = output #4
   endif
   if NCHAN >5
trisval &=	B'01111111'	;RA7 = output #5
   endif
   if NCHAN >6
trisval &=	B'11111110'	;RA0 = output #6
   endif
   if NCHAN >7
trisval &=	B'11111101'	;RA1 = output #7
   endif
	movlw	low trisval
	movwf	TRISA

trisval	set	~AUXOUTB & B'11101111'	;RB4 = output #0
   if NCHAN >1
trisval &=	B'11001111'	;RB5 = output #1
   endif
   if NCHAN >2
trisval &=	B'10001111'	;RB6 = output #2
   endif
   if NCHAN >3
trisval &=	B'00001111'	;RB7 = output #3
   endif
   if RCCHAN
trisval |=	B'00000001'	;RB0 = input(RC pulse)
   endif
 	movlw	low trisval
	movwf	TRISB
  endif
;***********************************************************

;--------------- Fill entire RAM with zeros:
	clrf	FSR0L
	movlw	0x20
	movwf	FSR0H
initclr				;clr all vars (memory range 0x20..0x7F, all banks)	
	clrf	INDF0
	incfsz	FSR0L,F		;address++
	bra	$+2
	incf	FSR0H,F
	bsf	INDF0,0		;check for RAM @selected address
	btfsc	INDF0,0		;it's not RAM, reached end of RAM 
	bra	initclr		;clr memory, set next address

;--------------- Initialize PWM engine
	call	initPWM
		 
	banksel	0		;setup registers in bank 0:
;--------------- Setup Timer 2
;TMR2 is used to generate Millisecond time base.
	movlw	b'00000111'	;T = 8s, timer = on
	movwf	T2CON
	movlw	124
	movwf	PR2		;interrupt every millisecond

;--------------- Setup Timer 1
;TMR1 is used for RC input pulse measuring

	movlw	b'00110001'	;T1 counting @1MHz
	movwf	T1CON
	clrf	TMR1L
	clrf	TMR1H

;--------------- Initialize ALIC instruction pointers
	movlw	NCHAN-1
	movwf	channel		;initialize ALIC IP for all channels
inipcplp	
	call	channelFSR0	;FSR0 -> channel struct
	call	pcp_reset	;set IP to start of program
	movf	ip_low,W
	movwi	c_IPL[FSR0]	;store IP in channel struct
	movf	ip_high,W
	movwi	c_IPH[FSR0]
	clrf	INDF0		;(clr JUMP flag: initial reset is not a jump)
	decf	channel,F
	btfss	channel,7
	bra	inipcplp	;>= 0: next channel

;---------------- enable interrupts
	banksel	PIE1
	bsf	PIE1,TMR2IE	;enable TMR2 interrupt (systick)
	banksel	0
  if RCCHAN
	call	rc_init		;initialize PPM input
	movlw	B'11101000'	;enable T0- and port-change interrupt
  else
	movlw	B'11100000'	;enable T0-int (PWM)
  endif
	movwf	INTCON	

#ifdef DEBUG_UART
	extern	init_UART	;(debug monitor)
	call	init_UART
#endif

;______________________________________________________________
; 	 		  			Infinite Mainloop							
;______________________________________________________________
mainlp
 if RCCHAN
	call	procRCIn	;call rc-service routine 
 endif
	banksel	0
	movf	tickcnt,F
	btfsc	STATUS,Z	;tickcnt > 0?
	bra	mainlp		;== 0: actually no ticks to process

;---------- time went forward by 1 systick (i.e. 1 Millisecond):
	decf	tickcnt,F	;one tick less remaining to process
	movlw	NCHAN-1
	movwf	channel		;process tick for each channel

	movlw	low (ChProcVars + (NCHAN-1)*PV_SIZE)
	movwf	FSR0L
	movlw	high (ChProcVars + (NCHAN-1)*PV_SIZE)
	movwf	FSR0H

;----------- loop for each output channel:
;(FSR0 always points to channel structure)  
mlpch
	moviw	c_msCnt[FSR0]
	decfsz	WREG,W		;decr. c_msCnt  
	bra	ld_msCnt	;no timeout yet

	moviw	c_IPL[FSR0]
	movwf	ip_low
	moviw	c_IPH[FSR0]
	movwf	ip_high
	call	chantick	;process tick(channel)
	movf	ip_low,W
	movwi	c_IPL[FSR0]	;adjust channel's IP
	movf	ip_high,W
	movwi	c_IPH[FSR0]

	btfsc	INDF0,JMPFLG
	bra	skip_ld_msCnt	;skip for wait/jump instr.

	movlw	OPC_DEFTICKLEN
	call	getparm		;get length of tick (channel)
	btfsc	STATUS,C	
	movlw	10		;parameter not present: default=10ms
ld_msCnt
	movwi	c_msCnt[FSR0]	;(re)load msCnt(channel)

skip_ld_msCnt
	bcf	INDF0,JMPFLG
	moviw	c_tau[FSR0]	;exponential decay active?
	btfsc	STATUS,Z
	bra	mlpchnxt	;no, process next channel

	moviw	c_tauCnt[FSR0]
	addlw	-1
	movwi	c_tauCnt[FSR0]
	btfsc	STATUS,Z	
	call	process_tau	

mlpchnxt
	addfsr	FSR0,-PV_SIZE	;-> next lower channel's vars
	decf	channel,F	;next channel...
	btfss	channel,7
	bra	mlpch		;channel >= 0

	bra	mainlp		;all channels done


;=============== exponential decay processing subroutine
process_tau
	moviw	c_pwdL[FSR0]	
	movwf	w16		;final pwm value -> w16
	moviw	c_pwdH[FSR0]
	movwf	w16+1
	
	moviw	c_pwoL[FSR0]	
	movwf	y_low		;y = current output value
	subwf	w16,F		;w16 = final val. - current val. 
	moviw	c_pwoH[FSR0]	
	movwf	y_high
	subwfb	w16+1,F		;C = 0 if result <0
	rrf	temp+1,F	
	comf	temp+1,F	;bit 7 = sign bit (1: negative)

	moviw	c_tau[FSR0]
	andlw	0x8F		;get value for reloading timer...
	iorlw	0x10		;default: T=16..31 ms
	btfss	WREG,7		;if MSB=1 -> time step *= 4
	bra	$+3		;MSB=0 -> skip shift ops
	aslf	WREG,W		;64..124 ms (4ms resolution)
	aslf	WREG,W
	movwi	c_tauCnt[FSR0]	;= time step (ms)

	movf	w16,W
	iorwf	w16+1,W		;actual difference == 0?
	btfsc	STATUS,Z
	return			;y, pwm == destination value, no further change
	
	moviw	c_tau[FSR0]	;b.4..6 = number of divisions (/2)
	swapf	WREG,W		;1..7 divisions -> /2../128
	andlw	0x07		;divide (W) times by 2 
	movwf	temp
	btfss	STATUS,Z
	bra	ptauLp
	
	moviw	c_tauCnt[FSR0]	;= time step (ms)
	bcf	WREG,4		;extreme little tau -> steptime < 16ms
	movwi	c_tauCnt[FSR0]	;time step (ms): < 16
	incf	temp,F		;still divide difference by 2!

ptauLp
	aslf	temp+1,W	;preserve MSB = sign bit
	rrf	w16+1,F		;pwm difference>>W 
	rrf	w16,F
	decfsz	temp,F
	bra	ptauLp

	btfsc	STATUS,Z	;w16 == 0?
	incf	w16,F		;0 -> 1: keep min value >0

	movf	y_low,W
	addwf	w16,F		;adjust to new pwm value
	movf	y_high,W
	addwfc	w16+1,F

;================== Adjust PWM, restore FSR0
Adjust_PW
	movf	w16,W
	movwi	c_pwoL[FSR0]
	movf	w16+1,W
	movwi	c_pwoH[FSR0]
	call	AdjustPW

;================== &processing vars(channel) -> FSR0
channelFSR0
	movf	channel,W
	call	mulPV_SIZE
	addlw	low ChProcVars
	movwf	FSR0L
	movlw	high ChProcVars
	btfsc	STATUS,C
	addlw	1
	movwf	FSR0H
	return

;================== returns W * PV_SIZE
mulPV_SIZE
	brw
  variable i=0
  while i<NCHAN	
	retlw	i*PV_SIZE					
    	variable i=i+1
  endw



;=============== calculate c_tau value from parameter word @IP
pcp_deftau
	call	read_ipword
calc_tau
	movlw	4		;default: 16 steps/tau
	movwf	temp		;number of divisions

	movf	w16+1,F
	btfsc	STATUS,Z	
	bra	calc_stepsby2	;tau < 256
	bra	incsteps

incsteps_lp
	rrf	w16,F		;time/2
	incf	temp,F		;double shifts & number of steps 
incsteps	
	lsrf	w16+1,F		;time/16 still > 32ms?
	btfss	STATUS,Z
	bra	incsteps_lp	;y: half time, double steps

	movlw	8
	subwf	temp,W		;> 7 shifts = 128 steps? (tau > 4s)
	btfss	STATUS,C
	bra 	store_ctau

	movlw	0x0E
	addwf	temp,F		;-2, set bit 3 = flag: time*4
	movlw	0xF0
	andwf	temp,W
	btfsc	STATUS,Z
	bra 	store_ctau

	movlw	0xFF		;still to many shifts:
	movwf	temp		;biggest c_tau possible! (ca.16s) 
	bra 	store_ctau
	
calc_t1
	lsrf	w16+1,F		;w16 /16-> w16,temp
	rrf	w16,F
	rrf	temp,F
	decfsz	WREG,W
	bra	calc_t1

	movlw	16
	subwf	w16,W
	movlw	0
	subwfb	w16+1,W
	btfss	STATUS,C
	bra	calc_stepsby2	;(w/16) < 16ms/step, reduce steps

; handling of tau < 256 ms: less than 16 time steps per tau
stepsby2_lp
	aslf	w16,F		; ms * 2
	btfsc	STATUS,C
	bra	store_ctau
calc_stepsby2
	decfsz	temp,F		; #shifts--, double time per step  
	bra	stepsby2_lp

;tau is very short, just two steps, step time < 16 ms
	movlw	0x30		
	subwf	w16,F
	btfss	STATUS,C	; result < 3ms?
	clrf	w16		;y: disable exponential decay

store_ctau
	movf	w16,W
	andlw	0xF0		; b.4-7 = (b0..3) of step time (0b1xxxx ms)  
	iorwf	temp,W		; b.0-2 = number of right shifts
	swapf	WREG,W		; adjust to storage format
	movwi	c_tau[FSR0]	; store
	andlw	0x0F
	iorlw	0x10		; =16..31ms
	movwf	temp
	moviw	c_tauCnt[FSR0]
	subwf	temp,W
	btfsc	STATUS,C	;actual c_tauCnt > temp? 
	return			;nope

	movf	temp,W		;y: limit c_tauCnt to new period time
	movwi	c_tauCnt[FSR0]
	return
	


;================ processing of ALIC instruction (IP->Opcode)
processIP
	call	read_ip		;get opcode of next instruction
	movwf	opcode
	movwf	y_high		;instruction or brightness val. -> y_high
	addlw	0x40
	btfss	STATUS,C
	bra	nextbrght	;opcode <0xC0: next brightness value

;================ execution of PCP control instruction (W)
;for opcodes >= 0xC0 (on entry: W = opcode-0xC0)
ctrlinstr
	addlw	-0x10	
	btfsc	STATUS,C	;check for range of opcodes (C0..CF)
	bra	ctrlinstr2	;opcode >= 0xD0: type 2

	addlw	0x10		;restore for table access...
	brw
	goto	inc_ip		;0xC0: define ticklen (header, just skip) 
	goto	inc_ip		;0xC1: define channel brightness (header, just skip) 
	goto	inc_ip		;0xC2: define dedicated switch (header, just skip) 
	goto	inc_ip		;0xC3: define dedicated variable (header, just skip) 
	goto	pcp_deftau	;0xC4: define time constant for exp. decay
	return			;0xC5: (invalid)
	return			;0xC6: (invalid)
	return			;0xC7: (invalid)

	goto	pcp_dly		;0xC8: delay
	goto	pcp_jump	;0xC9: JUMP
	goto	pcp_smthoff	;0xCA: smooth mode off
	goto	pcp_smthon	;0xCB: smooth mode on
	goto	pcp_rndjmp	;0xCC: random jump
	goto	pcp_rnddly	;0xCD: random delay
	goto	pcp_rndbrt	;0xCE: random brightness
	goto	pcp_reset	;0xCF: (undefined)

ctrlinstr2
	lsrf	WREG,W
	addlw	(ctrlt1-ctrlt2)
	btfsc	STATUS,C
	goto	pcp_reset	;opcode > defined range
	addlw	(ctrlt2-ctrlt1)
	brw

ctrlt1
	goto	pcp_haltswoff	;0xD0/0xD1: halt while switch is off
	goto	pcp_haltswon	;0xD2/0xD3: halt while switch is on
	goto	pcp_jswoff	;0xD4/0xD5: jump if switch is off
	goto	pcp_jswon	;0xD6/0xD7: jump if switch is on
	goto	pcp_jvarb	;0xD8/0xD9: jump if var is below
	goto	pcp_jvarae	;0xDA/0xDB: jump if var is above or equal
	goto	pcp_clrsw	;0xDC/0xDD: clear control switch
	goto	pcp_setsw	;0xDE/0xDF: set ctrl.switch 
	goto	pcp_setbrvar	;0xE0/0xE1: Set Brightness from Variable
	goto	pcp_setvar	;0xE2/0xE3: Set Variable
	goto	pcp_copy	;0xE4/0xE5: copy variable
	goto	pcp_addcvar	;0xE6/0xE7: add Const. to variable
	goto	pcp_addvvar	;0xE8/0xE9: add var. to variable
	goto	pcp_mulvar	;0xEA/0xEB: multiply var.by srcvar
	goto	pcp_scaleup	;0xEC/0xED: scale up
	goto	pcp_scaledown	;0xEE/0xEF: scale down
	goto	pcp_rndvar	;0xF0/0xF1: random value -> var
	goto	pcp_clip	;0xF2/0xF3: clip contents of var
ctrlt2


;================= Step PCP(channel) forward by one tick
chantick
	moviw	c_TCnt[FSR0]	
	addlw	-1
	movwi	c_TCnt[FSR0]	;TCnt--
	btfsc	STATUS,Z
	bra	chantout	;tick counter reached zero

	btfss	INDF0,SMOOTHFLG	;smooth mode active?
	return			;no, task finished
	
;----------- waiting for timeout: process brightness-interpolation
procesSmooth
	moviw	c_TCnt[FSR0]	;remaining ticks till next operation
	movwf	x1		;-> x1 (>0)

	moviw	c_Period[FSR0]
	movwf	x2		;= length of current time interval

	movf	x1,W		;remaining ticks
	subwf	x2,W		;full interval - remaining ticks -> x_high
	movwf	x_high			 

	moviw 	c_val1L[FSR0]
	movwf	y1_low
	moviw 	c_val1H[FSR0]
	movwf	y1_high
	moviw 	c_val2L[FSR0]
	movwf	y2_low
	moviw 	c_val2H[FSR0]
	movwf	y2_high

	call	interpolate

;================ set PWM drive value
setpwmadj
	movf	y_high,W
	movwf	w16+1
	movwi	c_pwdH[FSR0]
	movf	y_low,W
	movwf	w16
	movwi	c_pwdL[FSR0]

	moviw	c_tau[FSR0]
	btfss	STATUS,Z	;physical output value by tau-processing?
	return			;y: process_tau will adjust PW
	goto	Adjust_PW	;no: adjust pulse width immediately 


;---------- timeout of tick-counter: execute next instruction(s)
chantout
	btfss	INDF0,SMOOTHFLG	;smooth mode active?
	bra	cto_skipsmooth	;no

;reached end of current smooth period:
;	bcf	INDF0,SMOOTHFLG	;terminate previous smooth time period
	moviw 	c_val2L[FSR0]	;final drive value at end of this period
	movwf	y_low		
	moviw 	c_val2H[FSR0]	
	movwf	y_high
	call	setpwmadj	;adjust drive value = y
	call	read_ip		;W = delay after current brightness operation
	call	prepSmth	;set parameters for next smooth period
	bra	chantck1

cto_skipsmooth
	call	processIP	;execute instruction @IP

chantck1
	moviw	c_TCnt[FSR0]	;delay(channel) == 0?
	btfsc	STATUS,Z
	bra	chantout	;no delay: next instruction immediately
	return			;>0: processing finished for this step


;set new pwm value (y) from brightnes (y_high):
nextbrght
	bsf	INDF0,SMOOTHEOP ;case smooth mode/scanning: set flag
	btfsc	INDF0,SMOOTHFLG	;smooth mode active (scanning!)?
	return			;y: exit here...

	call	calcPWM		;new pwm value (y_high) ->y
	call	setpwmadj	;set new drive value = y
	call	read_ip		;delay value -> W
	movwi	c_TCnt[FSR0]	;set new delay time

	btfss	INDF0,SMOOTH_EN	;prepare smooth mode if enabled
	return			;smooth disabled, job done

; Prepare interpolation, starting from current value (y)
; input W = delay time after current brightness event 
; IP -> next instruction
prepSmth
	bcf	INDF0,SMOOTHEOP ;start scanning: clr flag
	bsf	INDF0,SMOOTHFLG	;smooth is now active
	movwf	saveTCnt	;save delay here, in case of c_TCnt has to be restored
	movwi 	c_Period[FSR0]	;sums up delay time until next brightval

	movf	ip_low,W	;(-> next opcode)
	movwi	c_IPL[FSR0]	;save current instruction pointer
	movf	ip_high,W
	movwi	c_IPH[FSR0]
	movlw	5		;maximum # of instructions to scan until next brightness event
	movwf	temp+1		
	movf	y_low,W		
	movwi 	c_val1L[FSR0]	;current value = start value of next period
	movf	y_high,W
	movwi 	c_val1H[FSR0]
	bra	prepSmthLp1	

prepSmthLp
	moviw 	c_Period[FSR0]
	movwf	temp
	moviw 	c_TCnt[FSR0]
	addwf	temp,W
	movwi	c_Period[FSR0]	;c_Period += c_TCnt
	btfsc	STATUS,C
	bra	CancelSmooth 	;period > 255: invalid!

prepSmthLp1
	movlw	0
	movwi 	c_TCnt[FSR0]
	
	movf	ip_low,W
	movwi 	c_val2L[FSR0]	;save current IP here...
	movf	ip_high,W
	movwi 	c_val2H[FSR0]

	call	processIP	;next instruction
	btfss	INDF0,SMOOTH_EN
	bra	CancelSmooth 	;smooth mode has been disabled
	btfss	INDF0,SMOOTHEOP ;set if "nextbrght" has been called
	bra	InterpScanNext	;not yet end of period

;scan process found end of period:
;IP -> delay after new set brightness instruction
	moviw 	c_Period[FSR0]	;period length
	movwi	c_TCnt[FSR0]	;load delay timer

	call	calcPWM		;get pwm value from brightness (y_high)
	movf	y_low,W		
	movwi 	c_val2L[FSR0]	;y = value at end of time period
	movf	y_high,W
	movwi 	c_val2H[FSR0]	
	return

;scan next instruction:	
InterpScanNext
	moviw 	c_val2L[FSR0]	;check for HALT-condition 
	xorwf	ip_low,W
	movwf	temp
	moviw 	c_val2H[FSR0]
	xorwf	ip_high,W
	iorwf	temp,W
	btfsc	STATUS,Z	;Z=1 if IP is same value as before processIP
	bra	CancelSmooth

	decfsz	temp+1,F	;reached max. number of ctrl.events to scan?
	bra	prepSmthLp	;not yet
		
CancelSmooth
	moviw	c_IPL[FSR0]	;restore instruction pointer
	movwf	ip_low
	moviw	c_IPH[FSR0]
	movwf	ip_high
	movf 	saveTCnt,W
	movwi	c_TCnt[FSR0]	;restore delay timer
;CancelSmooth1
	bcf	INDF0,SMOOTHFLG	;smooth is inactive for this code section
	return


;================ calculate pwm of (channel) from brightness value (y_high)
;input value of y_high=0x80 means 100%, will then be adjusted by
;brightness adjustment value from PCP-header (0x80 means * 1.0)   
calcPWM
	clrf	x_low
	clrf	y_low
	movf	y_high,W
	movwf	x_high
 #if GAMMACORRECT
	call	multiply	;y*y -> y (physiological brightness)
 #endif
	movlw	low (PWM_PERIOD*(1<<PWMSCALE))	
	movwf	x_low
	movlw	high (PWM_PERIOD*(1<<PWMSCALE))
	movwf	x_high
	call	multiply	;adjust to PWM value (left aligned!)
				;adjust by adj.value for this channel:
	movlw	OPC_DEFBRIGHT	;opcode: DEFBRIGHT
	call	getparm
	movwf	x_high		;(0x80 means *1.0)
	clrf	x_low
	btfsc	STATUS,C	;case such data in header: -> multiply
	return

;===================== Multiplication subroutine
;multiplies y * x -> y
;x,y = 16 bit fixpoint numbers (1.15), 0x8000 represents value 1.0

multiply
	movlw	16
	movwf	temp		;loop counter
	movf	y_high,W
	movwf	w16+1
	movf	y_low,W
	movwf	w16
	clrf	y_low
	clrf	y_high

	movf	x_low,W
	btfss	STATUS,Z
	bra	mul_2		;full 16 bit processing

	movf	x_high,W
	movwf	x_low
	lsrf	temp,F		;= 8 (8 bit processing)
	bra	mul_2

mul_lp
	rrf	y_high,F
	rrf	y_low,F		;result /2

mul_2
	lsrf	x_high,F
	rrf	x_low,F		;next bit of multiplier -> C
	btfss	STATUS,C
	bra	mul_lp_sk

	movf	w16,W
	addwf	y_low,F		;bit was 1: add multiplicant to result
	movf	w16+1,W
	addwfc	y_high,F

mul_lp_sk
	decfsz	temp,F
	bra	mul_lp		;next loop

	btfss	STATUS,C
	return

	movlw	0xFF		;overflow: set result to max.value
	movwf	y_high
	movwf	y_low
	return


;================ Linear Interpolation
;computes Y-value for given X-val. 
;input:
;y1 = Y-value for X=0
;y2 = Y-value for X=x2
;x2 = X-value: length of interval
;x = x input value (1 ... x2-1)
;output:
;y = Y-value for given x

interpolate
	banksel	x1
	clrf	x1
	clrf	temp
	btfsc	x2,7
	bra	interp_loop	;x2 >= 0x80

	bcf	STATUS,C
interp_max				
	rlf	x2,F		;maximize for higher precision
	rlf	x_high,F
	btfss	x2,7
	bra 	interp_max

interp_loop
	movf	y1_low,W
	addwf	y2_low,W
	movwf	y_low
	movf	y1_high,W
	addwfc	y2_high,W
	movwf	y_high
	rrf	y_high,F	;y = mean(y1, y2)
	rrf	y_low,F

	movf	x1,W
	addwf	x2,W
	rrf	WREG,W
	xorwf	temp,F		;Z=1 if there was no change since last loop
	btfsc	STATUS,Z
	return			;same mean value as last time: iteration complete
	movwf	temp		;temp = mean(x1, x2)
	subwf	x_high,W	; C=0 if x_high < mean(x1, x2)
	btfsc	STATUS,Z
	return			;x_high == mean(x1,x2) -> iteration complete

	movf	temp,W		;mean(x1,x2) -> W
	btfss	STATUS,C
	bra	interp_less	;C=0 => x_high < mean(x1,x2) 
;x_high > mean(x1,x2):
	movwf	x1		;new interval = mean..x2
	movf	y_low,W
	movwf	y1_low		;new y within y..y2
	movf	y_high,W
	movwf	y1_high
	bra	interp_loop

;x_high < mean(x1,x2):
interp_less
	movwf	x2		;new interval = x1..mean
	movf	y_low,W
	movwf	y2_low		;new y within y1..y
	movf	y_high,W
	movwf	y2_high
	bra	interp_loop


;=================== Read byte (pcpip(channel)) from memory
;interprets parameter as variable address on values > 0xF0 and
;returns data from variable, rather than constant parameter value.
read_ip_vr
	call	read_ip
	addlw	0x10		;parameter value 0xF0..0xFF: var_number->W
	btfsc	STATUS,C
	bra	var_sel1	;return contents of variable (W)

	addlw	0xF0		;parameter < 0xF0: return parameter value
	return
 
;=================== Read byte (ip++) from program memory
read_ip
	movf	ip_low,W
	movwf	FSR1L
	movf	ip_high,W	; (MSB=1 -> program flash)
	movwf	FSR1H
	movf	INDF1,W		;byte @ip -> W
inc_ip
	incfsz	ip_low,F	;ip++
	return
	incf	ip_high,F
	return

;__________________________________________________________
; Read word(ip++) -> w16
read_ipword
	call	read_ip		;ip->FSR1, ip++
read_FSR1word
 errorlevel-302
	banksel	EEADRL
	movf	FSR1L,W
	movwf	EEADRL
	movf	FSR1H,W
	movwf	EEADRH

	bcf	EECON1,CFGS 	; Do not select Configuration Space
	bsf 	EECON1,EEPGD 	; Select Program Memory
	bcf 	INTCON,GIE 	; Disable interrupts
	bsf 	EECON1,RD 	; Initiate read
	nop		
	nop
	bsf 	INTCON,GIE 	; Restore interrupts
	movf	EEDATH,W 	; Get MSB of word
	movwf	w16+1
	movf	EEDATL,W
	movwf	w16
	banksel	0
	return
 errorlevel+302

;================== get parameter from PCP header(channel)
;input: W = opcode for selection of parameter
;output W = parameter value, C=0
;	or C=1 if requested parameter was not found in header data
getparm
	movwf	temp		;save opcode to search for
	movf	channel,W
	addlw	low ALICFLASH	; program entry address -> w16
	movwf	FSR1L
	movlw	high ALICFLASH
	movwf	FSR1H
	call	read_FSR1word

	movwf	FSR1L
	movf	w16+1,W 	; Get MSB of word
	iorlw	0x80		; indirect access to FLASH 
	movwf	FSR1H

getprmnext
	movf	INDF1,W
	xorwf	temp,W
	btfsc	STATUS,Z
	bra	getprm_match	;found requested opcode

	addfsr	FSR1,2		;skip instruction (opcode+parameter)
	andlw	0xF8		;mask out all header-instr.
	btfsc	STATUS,Z
	bra	getprmnext

	bsf	STATUS,C	;(flag: requested parameter not found)
	return			;some other instr. but header

getprm_match
	moviw	++FSR1		;read parameter value
	addlw	0x10		;data range 0xF0..0xFF: access variable:
	btfsc	STATUS,C
	bra	var_sel1	;get data from variable

	addlw	0xF0		;restore original value
	bcf	STATUS,C	;return "W=parameter"
	return			;< 0xF0: use value directly (C=0)
			
 errorlevel+302

;********************************************************
;**** Subroutines for PCP control code processing    ****

;__________________________________________________________
pcp_reset			;P_RESET: Reset pcp(channel) to startadr
	btfsc	INDF0,SMOOTHFLG
	bra	pcp_smthoff	;smooth mode (scanning)

	movlw	low ALICFLASH
	addwf	channel,W
	movwf	ip_low		;IP -> ALIC reset address
	movlw	(high ALICFLASH)+0x80
	movwf	ip_high
	call	pcp_jump	;fetch reset address and load it into IP

	clrf	y_high
	clrf	y_low
	call	setpwmadj	;set output level = off
;__________________________________________________________
pcp_smthoff			;SMOOTHOFF: smooth mode off
	bcf	INDF0,SMOOTH_EN
	return
;__________________________________________________________
pcp_smthon			;SMOOTHON: smooth mode on
	bsf	INDF0,SMOOTH_EN
	return
;__________________________________________________________
pcp_jvarb			;jump if var < const.value
	call	mvar_sel
	btfsc	STATUS,C
	bra	pcp_skip2	;invalid var-adress: skip value, no jump

	movwf	temp		;cont. of variable
	call	read_ip		;get argument = value to compare with
	subwf	temp,W		;var-arg: C=0 if var<arg
	btfsc	STATUS,C
	bra	inc_ip		;arg >= var: no jump, skip
	bra	pcp_jump	;do the jump

;__________________________________________________________
pcp_jvarae			;jump if var >= const.value
	call	mvar_sel
	btfsc	STATUS,C
	bra	pcp_skip2	;invalid var-adress: skip value, no jump

	movwf	temp		;cont. of variable
	call	read_ip		;get argument = value to compare with
	subwf	temp,W		;var-arg: C=0 if var<arg
	btfsc	STATUS,C
	bra	pcp_jump	;do the jump
	bra	inc_ip		;arg < var: no jump, skip

;======= skip 2 arguments:
pcp_skip2
	movlw	2
	addwf	ip_low,F
	btfsc	STATUS,C
	incf	ip_high,F
	return

;__________________________________________________________
pcp_haltswon			;halt while switch or switch x is on	
	call	msw_sel		;select bit
	btfsc	STATUS,C
	return			;invalid switch: no halt

	andwf	INDF1,W
	btfsc	STATUS,Z
	return			;bit is clr: no halt
	bra	halt1

;__________________________________________________________
pcp_haltswoff			;halt while dedicated switch or switch x is off	
	call	msw_sel		;select switch bit
	btfsc	STATUS,C
	return			;invalid switch: no halt

	andwf	INDF1,W
	btfss	STATUS,Z
	return			;bit is set: no halt
halt1
	movlw	-1
	btfsc	opcode,ADRMODE	
	movlw	-2		;switch-# is operand: backup by two
	addwf	ip_low,F	;restore IP to "halt" instr.
	btfss	STATUS,C
	decf	ip_high,F

set_tickcnt_1
	movlw	1
	movwi	c_TCnt[FSR0]	;wait one tick before next action
	movlw	1		;1 ms until next operation, 
	movwi	c_msCnt[FSR0]	;regardless of ticktime-setting!
	bsf	c_Flags,JMPFLG
	return
;__________________________________________________________
pcp_jump			;JUMP
	btfss	INDF0,SMOOTHFLG	;skip if smooth mode active (scanning!)
	call	set_tickcnt_1	;avoid deadlock: continue on next tick	
	call	read_ipword	;get destination for jump (= word!)
	movwf	ip_low		; set new IP value...
	movf	w16+1,W
	iorlw	0x80		; MSB(address) = 1 for addressing through FSR
	movwf	ip_high		
 errorlevel+302
	return
;__________________________________________________________
pcp_rndjmp			;random jump
	call	rnd8
	andlw	0x7F
	movwf	temp		;save 7 bit random number
	call	read_ip_vr	;get "chance"
	subwf	temp,W
	btfss	STATUS,C
	bra	pcp_jump	;rnd7 < chance: execute jump
	bra	inc_ip
;__________________________________________________________
pcp_jswoff			;JSWOFF: jump if switch is off
	call	msw_sel		;select bit: dedicated channel switch
	btfsc	STATUS,C
	bra	pcp_jump	;C=1: invalid switch#, treat as "off"!

	andwf	INDF1,W
	btfsc	STATUS,Z
	bra	pcp_jump	;switch is off: execute jump
	bra	inc_ip		;condition = false: don't execute jump
	
;__________________________________________________________
pcp_jswon			;JSWON: jump if switch is on
	call	msw_sel		;select bit: dedicated channel switch
	btfsc	STATUS,C
	bra	inc_ip		;C=1: invalid switch#!

	andwf	INDF1,W	
	btfss	STATUS,Z
	bra	pcp_jump	;switch is on: execute jump
	bra	inc_ip		;off: skip
;__________________________________________________________
pcp_clrsw			;clear control switch:
	call	msw_sel		;select bit
	btfsc	STATUS,C
	return			;C=1: invalid switch#!

	xorlw	0xFF
	andwf	INDF1,F		;reset bit
	return
;__________________________________________________________
pcp_setsw			;set control switch:
	call	msw_sel		;select bit
	btfsc	STATUS,C
	return			;C=1: invalid switch#!

	iorwf	INDF1,F		;set bit
	return
	

;================= select dedicated switch (channel) or switch (read_ip),
;depending on ADRMODE-bit:
msw_sel
	btfsc	opcode,ADRMODE
	bra	switch_sel	;use switch number from parameter

;================= select dedicated switch (channel)
;returned: FSR1 = RAM address, W = bitmask
;C=1 if address is invalid
	movlw	OPC_DEFSWITCH
	call	getparm		;switch defined in header?
	btfsc	STATUS,C	;C=0: yes, W = switch#
	movf	channel,W	;C=1: no, use channel#
	bra	switch_s1

;================= select switch (read_ip)
;returned: FSR1 = address of register, W = bitmask
;C=1 if address is invalid
;
;map of switches:
;0..7 = default channel dedicated switches
;8..11 = RC memory switches, RC-channel 1
;12..15 = RC memory switches, RC-channel 2
;16..23 = I/O port A
;24..31 = I/O port B, C or GPIO, respectively
;32..47 = (reserved: RC Multiswitch decoder)
;48: RC input pulse detection (1: no valid RC pulses)

switch_sel
	call	read_ip	
switch_s1
	clrf	FSR1H		;main variables & PORTx are in bank 0!
	movwf	temp
	addlw	-16		;number < 16?
	btfsc	STATUS,C
	bra	switch_port	;>= 16: i/o-port access

	movlw	low switch	;0..7: select default switches
	btfss	temp,3
	bra	sw_fsr
 if RCCHAN
	movlw	high memswtch	;>= 8: select rc memory switch
	movwf	FSR1H
	movlw	low memswtch	;>= 8: select rc memory switch
	bra	sw_fsr
 else
	bsf	STATUS,C
	return			;invalid (C=1)!
 endif

switch_port
	addlw	-16
 if RCCHAN
	btfss	STATUS,C
	bra	range_port_sw	;<32: in range of port lines

	addlw	-16
	btfss	STATUS,C
	bra	range_ms	;<48: in range of multiswitch

	addlw	-1
	btfsc	STATUS,C
	return			;>49: invalid switch number (C=1)
	
	movlw	high rcdetect
	movwf	FSR1H
	movlw	low rcdetect	;>=0 if valid rc pulses detected
	movwf	FSR1L
	btfss	STATUS,Z	;Z=1 if SW=49
	retlw	B'00001000'	;mask bit 3 (RX0)
	retlw	B'10000000'	;mask MSB (RX1)	
	
 else
	btfsc	STATUS,C
	return			;>=32: invalid (C=1)!
 endif
range_port_sw
	movlw	low PORT_CH0	;= PORTB/PORTC/GPIO, depending on used MPU
 ifdef PINS_8
	btfsc	temp,3	
	bra	sw_fsr		;ok, 24..31 = PORTA
	bsf	STATUS,C	;invalid switch number!
	return
 else
	btfss	temp,3
	movlw	low PORTA	;16..23: port A
 endif
sw_fsr	movwf	FSR1L
	movf	temp,W
	andlw	0x07		;(for sure...)
	brw
	retlw	1
	retlw	2
	retlw	4
	retlw	8
	retlw	16
	retlw	32
	retlw	64
	retlw	128


;--------- switch number = 32..47, e.g. Multiswitch
range_ms
	bsf	STATUS,C	;not implemented here!
	return


;================= reads constant values (x,y) from parameters, 
;then selects variable depending on ADRMODE-bit, var contents -> W:
read_ip_xyvar
	call	read_ip_vr	;get min val
	movwf	x_high
	clrf	x_low
	call	read_ip_vr	;get max val
	movwf	y_high
	clrf	y_low

;================= selects channel's dedicated variable or variable from parameters,
;depending on ADRMODE-bit, var contents -> W:

mvar_sel
	btfsc	opcode,ADRMODE
	bra	var_sel		;ADRMODE=1: var. address from parameter (IP)

;================= select channel's dedicated variable
;var contents -> W
dvar_sel
	movlw	OPC_DEFVAR	
	call	getparm		;var defined in header?
	btfsc	STATUS,C	;C=0: yes, W = var#
	movf	channel,W	;C=1: no, use channel#
	bra	var_sel1

;================= select variable (read_ip)
;returned: FSR1 = address of register, W=data read from variable
;C=1, W=0 if address is invalid
;
;special map of registers:
;RCVAR: data of RC-channel 1
;RCVAR+1: data of RC-channel 2

var_sel
	call	read_ip
var_sel1
 	movwf	FSR1L		;= variable (x)
   if RCCHAN>0
     if RCCHAN>1
	addlw	-(RCVAR+2)	;for valid RCVAR: -> -2 or -1 
	btfsc	STATUS,C
	bra	varsel_skipRC	;var-# is not RC-value (>RCVAR+1)
	addlw	2		;for valid RCVAR: -> 0 or 1 
	btfss	STATUS,C
	bra	varsel_skipRC	;var-# is not RC-value (<RCVAR)

	addlw	low rcval
     else
	addlw	-RCVAR		;for valid RCVAR: -> 0 
	btfss	STATUS,Z
	bra	varsel_skipRC	;var-# is not RC-value

	movlw	low rcval
     endif
	movwf	FSR1L
	movlw	high rcval
	movwf	FSR1H
	bcf	STATUS,C
	movf	INDF1,W
	return

varsel_skipRC
	movf	FSR1L,W
   endif
 if NVARS > 0
	addlw	-NVARS
	btfsc	STATUS,C
	retlw	0		;invalid variable number! (C=1)

	movlw	low varbase
	addwf	FSR1L,F		;= address of variable (x)
	movlw	high varbase
	movwf	FSR1H
	movf	INDF1,W		;(C=0 by prec. addition)
	return
 else
	bsf	STATUS,C	;return C=1 if there are no variables
	retlw	0
 endif


;__________________________________________________________
pcp_dly				;Delay:
	call	read_ip_vr	;get delay value
	movwi	c_TCnt[FSR0]	;set new delay time
	return

;======================= generate next 8 bit random number
rnd8
	lsrf	random,W
	btfsc	STATUS,C
	xorlw	0xB4
	addwf	TMR0,W
	movwf	random
	return

;__________________________________________________________
pcp_rnddly			;Random Delay:
	call	read_ip		;get val for max. delay
	movwf	x_high
	call	rnd8		;new 8 bit random value
	addlw	1
	movwf	y_high
	rrf	y_high,F	;y = random number 0.0 ... 1.0
	call	multiply
	movf	y_high,W
	movwi	c_TCnt[FSR0]	;set new delay time
	return

;__________________________________________________________
pcp_rndbrt			;set Random Brightness:
	call	read_ip_vr	;get min. brightness
	movwf	x_high
	call	read_ip_vr	;get max. brightness 
	movwf	y_high

	call	rndxy
	goto	nextbrght	;sets brght=Y, next delay


;__________________________________________________________
pcp_rndvar			;load random value into variable:
	call	read_ip_xyvar	;get min, max and variable
	btfsc	STATUS,C
	return			;invalid var-address: return

	call	rndxy
	bra	styindf		;y -> INDF1

;__________________________________________________________
pcp_setbrvar			;set brightness = contents of variable:
	call	mvar_sel	;get data from variable
	btfsc	STATUS,C
	movlw	0		;invalid var: 0
	movwf	y_high
	goto	nextbrght	;sets brght=Y, next delay

;__________________________________________________________
pcp_clip			;clip contents of variable:
	call	read_ip_xyvar	;get min, max and variable
	btfsc	STATUS,C
	return			;invalid var-adress: return

	movf	x_high,W	;min. value
	subwf	INDF1,F		;var-min
	btfss	STATUS,C
	bra	stwindf		;var<min: W->var

	addwf	INDF1,F		;restore original value
	movf	y_high,W	;max. value
	subwf	INDF1,F		;var-max
	btfss	STATUS,C
	addwf	INDF1,W		;var < max: original value -> W
stwindf
	movwf	INDF1
	return

;__________________________________________________________
pcp_setvar			;set contents of variable:
	call	read_ip		;get value
	movwf	x_high

	call	mvar_sel	;select variable
	btfsc	STATUS,C
	return			;invalid var-adress: return
stxindf
	movf	x_high,W
	movwf	INDF1
	return

;__________________________________________________________
pcp_copy			;copy contents of variable:

	call	var_sel		;select source variable
	movwf	x_high

	call	mvar_sel	;select dest.-variable
	btfsc	STATUS,C
	return			;invalid var-adress: return

	bra	stxindf		;x_high -> INDF1, return

;__________________________________________________________
pcp_mulvar			;multiply variable by variable
	call	var_sel		;get multiplier
	movwf	x_high
	call	mvar_sel	;select variable
	btfsc	STATUS,C
	return			;invalid var-address: return

	movwf	y_high
	call	multiply
	bra	styindf		;result (y) -> var.

;__________________________________________________________
pcp_addvvar			;add variable to variable
	call	var_sel
	bra	addw2var

;__________________________________________________________
pcp_addcvar			;add constant value to variable

	call	read_ip		;get value
addw2var
	movwf	x_high

	call	mvar_sel	;select variable
	btfsc	STATUS,C
	return			;invalid var-adress: return

	clrf	temp		;temporary higher byte
	btfsc	x_high,7
	decf	temp,F		;expand sign bit

	movf	x_high,W
	addwf	INDF1,F
	btfsc	STATUS,C
	incf	temp,F

	movlw	0
	btfsc	temp,7
	bra	stwindf		;result <0: store 0

	movlw	255
	movf	temp,F		
	btfss	STATUS,Z
	bra	stwindf		;result >256: store 255

	return

;__________________________________________________________
pcp_scaleup			;scale variable up from given range to full range

	call	read_ip_xyvar	;get first/second limit -> x,y, select variable
	btfsc	STATUS,C
	return			;invalid var-address: return

;prepare data for interpolation:
;y1 = Y-value for X <= lower limit
;y2 = Y-value for X >= upper limit
;x2 = X-value: length of imput interval
;x = input value (offset from lower limit, 1 ... x2-1)
;interpolation computes:
;y = Y-value for given x

	clrf	y1_low
	clrf	y2_low
	clrf	y1_high
	clrf	y2_high

	movf	x_high,W
	subwf	y_high,W	;C=0 if x > y 
	btfsc	STATUS,C
	bra	scale_noswap	;y >= x

	movf	x_high,W	;swap x <-> y
	SWAPWF	y_high
	movwf	x_high
	subwf	y_high,W	;W=length of input interval (C=1)
	rrf	y1_high,F	;=0x80, C=0, output interval 128..0

scale_noswap
	rrf	y2_high,F	;y1-y2 = output interval (0..128 or 128..0) 
	movwf	x2		;= length of input interval	
	movf	INDF1,W
	subwf	x_high,W	;lower limit - input -> W
	btfsc	STATUS,C
	bra	scexit_low	;input <= lower limit: return y = y1
	
	sublw	0		;negate: input - lower limit -> W
	movwf	x_high		;(x_high = offset to lower limit)
	subwf	x2,W		;upper limit - x -> W
	btfss	STATUS,C
	bra	scexit_high	;x > upper limit
	btfsc	STATUS,Z
	bra	scexit_high	;x == upper limit

	call	interpolate
	movf	y_high,W
scexit
	movwf	INDF1
	return

scexit_low
	movf	y1_high,W
	bra	scexit
scexit_high
	movf	y2_high,W
	bra	scexit


;__________________________________________________________
pcp_scaledown			;scale variable down to given range

	call	read_ip_xyvar	;get first/second limit -> x,y, select variable
	btfsc	STATUS,C
	return			;invalid var-address: return

	movf	INDF1,W
	call	downscale
styindf
	movf	y_high,W
	movwf	INDF1
	return


;__________________________________________________________
;subroutine: random number generation in range x..y
rndxy
	call	rnd8
	andlw	0x7F		;7 bit random value (fixpoint notation: <1)

;__________________________________________________________
;subroutine for downscaling
;input:
;W = data to be scaled
;x, y = scaling interval (min., max)  
;output:
;y = scaled value

downscale
	movwf	y1_high
	clrf	y1_low	
	movf	x_high,W
	movwf	x1		;(= interval, min.)
	subwf	y_high,F	;max - min -> y (=multiplier)

	movf	y1_high,W	;value to be scaled -> W
	btfsc	STATUS,C
	bra	downsc2		;max > min: I/O = same sign
	
	movf	y_high,W	;<0 
	sublw	0		;negate...
	movwf	y_high		;>0
	subwf	x1,F		;adjust to true min. of interval

	movf	y1_high,W
	sublw 	128		;swap min/max input value (0..128 -> 128..0)
	btfss	STATUS,C
	movlw	0		;was > 128: limit min. to 0
		
downsc2
	movwf	x_high
	call 	multiply
	movf	x1,W
	addwf	y_high,F	;y += interval min.
	return

ALIC_CODE code ALICFLASH
  variable i=0
  while i<NCHAN						
	dw	entry_ch#v(i)		
	variable i=i+1
  endw
#include ALICFILE

	end