#include picalic_config.inc
#include MCU_Headers_em.inc	;Pocessor specific definitions
#include macros.inc

#define SHORTPW 4		;special PWM processing if pulse is shorter
#define TMARGIN 6		;minimum time margin between PWM queue events 

 if NCHAN > (PWM_PERIOD/(8*TMARGIN))
	error "PWM_PERIOD value too low!"
 endif 

	extern 	temp,w16			;(udata_shr)
	extern	channel,tickcnt			;(bank0)

;global routines:
	global AdjustPW
	global initPWM

;global variables:
	global mscnt

TEMP0 	udata_ovr 0x20		

delta		res 2	
displace	res 2	
new_pw		res 2		;pulse width (s/16)

	udata_shr
mscnt 		res 2		;infinite milliseconds counter
tmr0h 		res 1		;extends TMR0
pwm_queue_nxt 	res 1

;================================================
;banked data section:
PWMDAT 	udata

;================================================
;lineary addressed data section for PWM-event queue: 
LINEAR0	udata
pwm_queue res 2*NCHAN*Q_SIZE	;PWM queue starts here
				;queue_data ("on"/"off" events) for each channel

;order of queue events in memory:
;ch0_on, ch0_off, ch1_on, ch1_on... chN_on, chN_off

;queue events data structure:
	CBLOCK 0		;(labels are an offset address!)
q_next				;pointer to next event to process, 0 = not linked
q_actn				;denotes function to execute for this event
q_timeL 			;time to next event in queue
q_timeH
Q_SIZE:0
	ENDC


 errorlevel-302

;=============================== Interrupt service routine

INTERRUPT code 0x04

interrupt
#ifdef DEBUG_UART		;(adds exactly 1s of latency)
 	extern intTMR1, intIOC
	banksel	IOCAF
	movf	IOCAF,W		;change of input pin detected?
	btfss	STATUS,Z
	goto	intIOC

	banksel	0
	btfsc	PIR1,TMR1IF	;timer overflow of timer 1?
	goto	intTMR1		;y: UART I/O
	nop
#endif

	banksel	0
	pagesel 0
 if RCCHAN				;if RC-interface (input):
    extern rc_isr
	btfsc	INTCON,IOCIF		;interrupt from RC?
	goto	rc_isr			;process RC-interrupt	
 else
	nop				;without RC input:
	nop				;force the same timing
 endif
	btfsc	INTCON,TMR0IE		;if T0-int is disabled -> t0isr_skip
	btfss	INTCON,TMR0IF		;if T0-int is pending -> t0isr
	bra	t0isr_skip		;it's not T0-interrupt

;-------------- TMR0: PWM generator
;(T=X for reference)
t0isr
	bcf	INTCON,TMR0IF		;clr timer int.
	incfsz	tmr0h,F
	retfie				;not yet timeout

	movlw	high pwm_queue
	movwf	FSR0H			;use linear data memory
	movf	pwm_queue_nxt,W		;points to event to process now
	movwf	FSR0L
					;<== TMR0 counts here (Prescaler = 0)
	moviw	q_next[FSR0]
	movwf	pwm_queue_nxt		;-> next event in queue: to process next

	moviw	q_timeL[FSR0]
	incf	TMR0,F			;timing correction (delay by reset of prescaler and
	nop				;write TMR0 op.)
	subwf	TMR0,F
	moviw	q_timeH[FSR0]
	subwfb	tmr0h,F
	btfss	STATUS,C
	bra	t0isr_to1		;TMR0 adjusted to <0: next int on TMR0 overflow

	decf	tmr0h,F			;-> 0xFF
	bsf	INTCON,TMR0IF		;set TMR0 interrupt pending

t0isr_to1
	moviw	q_actn[FSR0]		;opcode -> W
	brw
;***********  while loops generate code for lookup table:
  variable i=0
  while i<NCHAN						
	bcf	PORT_CH#v(i),PIN_CH#v(i)	;opcode = 2*channel: output=low for selected channel
	retfie
	variable i=i+1
  endw

  variable i=0
  while i<NCHAN						
	bsf	PORT_CH#v(i),PIN_CH#v(i)	;opcode = 2*NCHAN + 2*channel: output=high for selected channel 
  	retfie
	variable i=i+1
  endw

  variable i=0
  while i<NCHAN					;opcode = 4*NCHAN + channel:	
	bra	PwmCh#v(i)ShortPulse		;pulse width < SHORTPW, process complete pulse here
  variable i=i+1
  endw

#if PWM_DITHER
  variable i=0
  while i<NCHAN					;opcode = 5*NCHAN + channel:	
	bra	PwmCh#v(i)DitherPulses		;pulse dithering, using 1s pulses
  variable i=i+1
  endw
#endif

;*********** (end of lookup table)

#if PWM_DITHER
 variable i=0
 while i<NCHAN				;while-loop generates code for each channel
PwmCh#v(i)DitherPulses
	call	dither_common
  if ACTIVELOW
	bcf	PORT_CH#v(i),PIN_CH#v(i)
  else
	bsf	PORT_CH#v(i),PIN_CH#v(i)
  endif
	call	dith_delay		;1s pulse
  if ACTIVELOW
	bsf	PORT_CH#v(i),PIN_CH#v(i)
  else
	bcf	PORT_CH#v(i),PIN_CH#v(i)
  endif
	retfie
 variable i=i+1
 endw

dith_delay
	bra	$+1
	nop
	return

dither_common
	moviw	q_timeH+Q_SIZE[FSR0]	;pulse counter
	addlw	1			;++
	movwi   q_timeH+Q_SIZE[FSR0]
	andlw	0x0F			;use only lower 4 bits
	movwf	FSR1L			;temporary storage
	moviw	q_timeL+Q_SIZE[FSR0]	;effective pulse width (s/16)
	subwf	FSR1L,W			;C=0 if counter < PW
	btfss	STATUS,C		;counter >= PW: skip this pulse
	return

	banksel	STKPTR
	decf	STKPTR,F		;discard return address
	retfie				

	 
#endif ;(PWM_DITHER)


;---------------------- PWM ISR processing: generate short pulse (used if PW < SHORTPW) 
  variable i=0
  while i<NCHAN					;while-loop generates code for each channel
PwmCh#v(i)ShortPulse
	
  if ACTIVELOW
	bcf	PORT_CH#v(i),PIN_CH#v(i)
  else
	bsf	PORT_CH#v(i),PIN_CH#v(i)
  endif
	call	delay_us		;delay time = q_timeL(ch.off-event)
  if ACTIVELOW
	bsf	PORT_CH#v(i),PIN_CH#v(i)
  else
	bcf	PORT_CH#v(i),PIN_CH#v(i)
  endif
	retfie
  variable i=i+1 
  endw

;=================== Short time delay loop for PW < SHORTPW
delay_us
	moviw	q_timeL+Q_SIZE[FSR0]	;delay time in s/2
	addlw	-3			
	btfss	STATUS,C		;<= 1s delay?	
	return				;y: "call delay_us" = total 7 cycles

delay_usLoop
	decf	WREG,W
	btfss	WREG,7
	bra	delay_usLoop
	return

t0isr_skip
	btfss	PIR1,TMR2IF
	retfie				;??? invalid interrupt source
;	bra	badinterrupt

;------------- TMR2: Increment milliseconds (system timer): 
	bcf	PIR1,TMR2IF
	incf	tickcnt,F
	incfsz	mscnt,F
	retfie
	incf	mscnt+1,F
	retfie

;badinterrupt
;	movlw	1<<2			;debug test!
;	xorwf	PORTA,F
;	retfie

;==============================================================================
;	Program code for initialization and for setting of PMW pulse width
PROG0 	code

;=========================== initialization:
initPWM
	movlw	NCHAN
	movwf	temp
	movlw	high pwm_queue
	movwf	FSR0H				;use linear data memory
	movlw	low pwm_queue
	movwf	FSR0L
	movwf	pwm_queue_nxt			;init pwm queueing
	decf	tmr0h,F				;= -1

;--------------- Setup Timer 0
	banksel	OPTION_REG
	movlw	b'00000010'			;PS = 1/8: Timer 0 counting microseconds 
	movwf	OPTION_REG

	banksel	TMR0
	clrf	TMR0

;	banksel channel				;(bank 0)
initpq1	
	aslf	channel,W			;0,2,4... = opcode for setting output low
 if ACTIVELOW
	addlw	2*NCHAN				;-> 2*NCHAN + 2*channel = opcode for setting output high
 endif
	movwi	q_actn[FSR0]
	movlw	low(PWM_PERIOD/NCHAN)		;distribute timeouts equally over full pwm period
	movwi	q_timeL[FSR0]
 	movlw	high(PWM_PERIOD/NCHAN)
	movwi	q_timeH[FSR0]
					
	incf	channel,F
	addfsr	FSR0,2*Q_SIZE			;set FSR0 to next queue element
	movf	FSR0L,W
	movwi	-2*Q_SIZE[FSR0]			;q_next(previous el.) -> next element 
	decfsz	temp,f				;finished queue[0..NCHAN-1]?
	bra	initpq1				;not yet...

	movlw	low pwm_queue
	movwi	-2*Q_SIZE[FSR0]			;q_next -> first element 

	return			


;=================== subroutine: turn output (channel) on 
ChOutON
	banksel PORTA
	aslf	channel,W
	brw
  variable i=0
  while i<NCHAN				;while-loop generates code for each channel
  if ACTIVELOW
	bcf	PORT_CH#v(i),PIN_CH#v(i)
  else
	bsf	PORT_CH#v(i),PIN_CH#v(i)
  endif
	return
  variable i=i+1 
  endw

StoreQActnOff
	movwi	q_actn[FSR1]		;store queue action (W), set output "off"

;=================== subroutine: turn output (channel) off 
ChOutOFF
	banksel PORTA
	aslf	channel,W
	brw
  variable i=0
  while i<NCHAN				;while-loop generates code for each channel
  if ACTIVELOW
	bsf	PORT_CH#v(i),PIN_CH#v(i)
  else
	bcf	PORT_CH#v(i),PIN_CH#v(i)
  endif
	return
  variable i=i+1 
  endw

;==================== set W = address(low) to access leading edge event (channel)
queueChW
 if Q_SIZE != 4
	ERROR change W*Q_SIZE calculation here!
 else
	swapf	channel,W		;channel * 2*Q_SIZE (=> * 8)
	lsrf	WREG,W
 endif
	addlw	low pwm_queue		;+ startaddress of queue
	return

;==================== set FSR1 register to access leading edge event (channel)
;(assumes that FSR1H was set to "high pwm_queue" already!)
;(returns W=FSR1L)
queueChFSR1
	call	queueChW
	movwf	FSR1L
	return

;==================== set FSR registers to access pwm-queue events
;input: W = 2*channel (pwm_on) or 2*channel+1 (pwm_off)
;output: FSR1 = requested queue event
;	 FSR0 = previous event in queue
;		(FSR0L = 0 if element is unlinked)
queueFSR
 if Q_SIZE != 4
	ERROR change W*Q_SIZE calculation here!
 else
	aslf	WREG,W			;W * Q_SIZE (=> W * 4)
	aslf	WREG,W
 endif
	addlw	low pwm_queue		;+ startaddress of queue
	movwf	FSR1L
	movwf	FSR0L
	movlw	high pwm_queue		;(linear adressing)
	movwf	FSR0H
	movwf	FSR1H
	
;search through whole queue until link to actual element is found:  
queueFSR_Srch
	moviw	q_next[FSR0]
	movwf	FSR0L
	btfsc	STATUS,Z
	return				;this element is not linked

	moviw	q_next[FSR0]
	subwf	FSR1L,W			;linked to current element?
	btfss	STATUS,Z
	bra	queueFSR_Srch		;no: try next...

	return

;============================================================== 
;Subroutine: set FSR registers for insertion of new event
;input: 
;	FSR1 = queue event at reference time (T=0)
;	w16 = rtime (s relative to ref.) for new event
;output: 
;	FSR0 = previous event (T < rtime)
;	FSR1 = next event (T >= rtime)
;	w16 = dtime (time between previous event and rtime)
;	delta = dtime (s between new event and next event in queue)
;	C = 1 if event does not collide with other events in queue 
;============================================================== 
queuePrepareInsert		
	movf	FSR1L,W
	movwf	FSR0L			;this is now "previous event"
	moviw	q_next[FSR1]
	movwf	FSR1L			;FSR1 -> next event

	moviw	q_timeL[FSR0]
	subwf	w16,F			;subtract dtime(event) 
	moviw	q_timeH[FSR0]
	subwfb	w16+1,F			;= time remaining to rtime
	btfsc	STATUS,C
	bra	queuePrepareInsert	;did not yet pass rtime

	clrf	delta
	clrf	delta+1
	movf	w16,W			;(w16 is negative here)
	subwf	delta,F
	movf	w16+1,W
	subwfb	delta+1,F		; delta = -w16  => delta >= 0
	
	moviw	q_timeL[FSR0]
	addwf	w16,F			;restore w16 = time between previous event and rtime
	moviw	q_timeH[FSR0]
	addwfc	w16+1,F

;check for sufficient time margin between PWM queue events: 
	call	queueChW		;address of event (channel, leading edge) -> W
	subwf	FSR0L,W
	btfsc	STATUS,Z
	bra	skipLM			;skip testing here if previous event is same PWM channel

	movlw	TMARGIN			
	subwf	w16,W
	movlw	0
	subwfb	w16+1,W
	btfss	STATUS,C
	return				;insufficient gap (previous<->new event): C=0

skipLM
	movlw	TMARGIN
	subwf	delta,W
	movlw	0
	subwfb	delta+1,W		;if insufficient gap (new<->following event: C=0
	return


;============================================================== 
;Subroutine: adjust PWM pulse width 
;input: 
;	channel: 	PWM channel selection
;	w16:		pulse width (s<<PWMSCALE)
;==============================================================
AdjustPW
	banksel	channel
 variable i=PWMSCALE-4
 while i>0				;PWMSCALE > 4:
	lsrf	w16+1,F			;scale to s/16
	rrf	w16,F
 variable i=i-1
 endw
	movm16	w16,new_pw
	aslf	channel,W
	addlw	1			;(-> trailing edge event)
	call	PqUnlink		;remove old trailing edge event from queue
	call	queueChFSR1		;FSR1 -> leading edge event
	movf	new_pw+1,W
	btfss	STATUS,Z
	bra	AdjustPW_nz1		;Pulse width >= 16s (MSB != 0)
	movf	new_pw,W
 if PWMSCALE >= 4	
	andlw	0xF0			;suppose < 1s = 0
 else
	andlw 	0xFF<<PWMSCALE
 endif 
	btfss	STATUS,Z
	bra	AdjustPW_nz		;new PW is not zero

#if PWM_DITHER
	movf	new_pw,W
	btfsc	STATUS,Z
	bra	AdjustPW_zero		;pw == 0
 variable i=4-PWMSCALE
 while i>0				;PWMSCALE < 4:
	aslf	WREG,W			;scale to s/16 in lower nibble
 variable i=i-1
 endw
	movwi	q_timeL+Q_SIZE[FSR1]	;store pulse width (s/16) 
	movf	channel,W
	addlw	5*NCHAN			;= opcode for dithering
	bra	StoreQActnOff
#endif

;--- new PW is zero: force pwm output to "off" state
AdjustPW_zero
	aslf	channel,W		;0,2,4... = opcode for output low
 if ACTIVELOW
	addlw	2*NCHAN			;->  opcode for output high if ACTIVELOW
 endif
	bra	StoreQActnOff

;--- new PW is 1..255:	
AdjustPW_nz
 if PWMSCALE >= 4	
	addlw	-SHORTPW<<4
 else
	addlw	-SHORTPW<<PWMSCALE
 endif 

	btfsc	STATUS,C
	bra	AdjustPW_nz1		;new PW > SHORTPW: separate events for on and off

;--- very short PW (< SHORTPW): only leading edge event is used, which then generates both edges  
  if PWMSCALE == 0
	aslf	new_pw,W
  endif
  if PWMSCALE == 1
	movf	new_pw,W
  endif
  if PWMSCALE == 2
	lsrf	new_pw,W
  endif
  if PWMSCALE == 3
	lsrf	new_pw,W
	lsrf	WREG,W	
  endif
  if PWMSCALE >= 4	
	lsrf	new_pw,W
	lsrf	WREG,W
	lsrf	WREG,W
  endif
		
	;W = pulse width (s/2)
	movwi	q_timeL+Q_SIZE[FSR1]	;store counter value for delay loop (delay_us) here

	movlw	4*NCHAN			;skip chx_low, chx_high opcodes
	addwf	channel,W		;-> pulse opcode (channel)
	bra	StoreQActnOff

;--- new PW >= SHORTPW:	
AdjustPW_nz1
 if PWMSCALE >= 2	
   if PWMSCALE >= 4	
	movlw	4
   else
	movlw	PWMSCALE
   endif
shifty_lp				
	lsrf	new_pw+1,F		;y/16 -> s units (20 cycles, 5 words
	rrf	new_pw,F
	decfsz	WREG,W
	bra	shifty_lp
 else
   if PWMSCALE == 1
	lsrf	new_pw+1,F	
	rrf	new_pw,F
   endif	
 endif

	aslf	channel,W		;0,2,4... = code for output low
 if ACTIVELOW==0
	addlw	NCHAN*2			; set opcode for output high
 endif
	movwi	q_actn[FSR1]		;set action (leading edge) = "PWM_ON"

	movlw	low (PWM_PERIOD-8)	;check for "near 100%" condition
	subwf	new_pw,W
	movlw	high (PWM_PERIOD-8)
	subwfb	new_pw+1,W
	btfsc	STATUS,C			
	return				;PW is nearly PWM_PERIOD or above: set 100% duty

;------------- PWM needs separate events for leading & trailing edge:
	clrf	displace+1			;displace = offset time to move both events
	clrf	displace
	movm16	new_pw,w16			;pulsewidth -> w16
	bra	AdjustPW_InsCheckNxt


;queuePrepareInsert found insufficient margin arround current insertion point.
AdjustPW_Displace
	call	SearchSlot		;Search for appropriate timing slot for trailing edge event

;displace-value denotes possible insertion point for trailing edge event.
;Check if leading edge event can be displaced by same amount of time:
	call	queueChFSR1		;FSR1 = leading edge event
	movf	displace,W		;displacement
	movwf	w16			;= new (preliminary) insertion point for this event
	movf	displace+1,W
	movwf	w16+1
	call 	queuePrepareInsert	;prepare data (leading edge) for SearchSlot
	btfss	STATUS,C		;if C=1: margins are ok, skip SearchSlot
	call	SearchSlot		;search a sufficient slot for leading edge
	
	movf	new_pw,W			;pulse width
	addwf	displace,W		;+ displacement
	movwf	w16			; = time between current leading edge and new trailing edge
	movf	new_pw+1,W
	addwfc	displace+1,W
	movwf	w16+1

	call	queueChFSR1		;FSR1 = origin (leading edge)

AdjustPW_InsCheckNxt
	call	queuePrepareInsert	;locate new event for trailing edge
	btfss	STATUS,C		;does it fit with proper margins?
	bra	AdjustPW_Displace	;doesn't fit, need (more) displacement!

;--------------- Insert new trailing edge event
	call	queueChW
	addlw	Q_SIZE			;-> trailing edge event
	movwf	FSR1L

	aslf	channel,W		;0,2,4... = code for output low
 if ACTIVELOW
	addlw	2*NCHAN			;->  code for output high
 endif
	movwi	q_actn[FSR1]
	call	PqLink			;insert it into the queue
	movf	displace,W
	iorwf	displace+1,W		;displacement > 0?
	btfss	STATUS,Z
	call	DisplaceLE		;y: displace leading edge event

;----------------- adjust current PWM output level (channel)
;this syncronizes the port's output state with PWM processing state in order to avoid
;false output states remaining for the rest of the current PWM period
	call	queueChFSR1		;FSR1 -> leading edge event
	addlw	Q_SIZE		
	movwf	FSR0L			;FSR0 -> trailing edge event

	bcf	INTCON,TMR0IE		;disable PWM Interrupt
	moviw	q_next[FSR0]
	subwf	pwm_queue_nxt,W		;next event == active ISR event?
	btfss	STATUS,Z
	bra	AdjO_lp			;no
	
	movf	TMR0,W
	btfsc	INTCON,TMR0IF		;did meanwhile timer overflow?
	movlw	0xFF			;y: use value just before overflow
	movwf	delta
	movf	tmr0h,W
	movwf	delta+1			;delta = -remaining time until next event

	moviw	q_timeL[FSR0]		
	addwf	delta,F
	moviw	q_timeH[FSR0]
	addwfc	delta+1,W		;is remaining time shorter than q_time?
	btfsc	STATUS,C
	bra	AdjO_off		;y: time is already beyond trailing edge -> off

	movwf	tmr0h			;set timer: time before trailing edge event 
	clrf	TMR0
	bcf	INTCON,TMR0IF
	movf	delta,W
	movwf	TMR0
	movf	FSR0L,W
	movwf	pwm_queue_nxt		;next evemt = trailing edge
	bra	AdjO_on
	

;check if current ISR-event is between leading and trailing edge.
;If true: current output state must be "on", else "off".
AdjO_lp
	moviw	q_next[FSR1]
	movwf	FSR1L
	subwf	pwm_queue_nxt,W		
	btfsc	STATUS,Z		;match with currently active ISR event?
	bra	AdjO_on			;y: -> output is now in on-state 

	movf	FSR1L,W
	subwf	FSR0L,W			
	btfss	STATUS,Z		;reached trailing edge event?
	bra	AdjO_lp			;no, check next event

AdjO_off
	call 	ChOutOFF
	bsf	INTCON,TMR0IE
	return

AdjO_on
	call	ChOutON
	bsf	INTCON,TMR0IE
	return

;-------------------- find next time gap with sufficient margins between
;preceding ("lower") and following ("upper") next events in queue
;input = result from queuePrepareInsert:
;	FSR1: next event, 
;	FSR0: previous event
;	x: time from current insertion point to next event
;	w16: time from previous event to current insertion point
;
;	displace = displacement time, sums up until possible insertion point has been found
;output:displace adjusted for new insertion point 
; 
SearchSlot
	call	queueChW
	subwf	FSR0L,W
	btfsc	STATUS,Z
	bra	SearchSlot_2		;skip lower margin test on own channel
	
	movf	w16+1,F			;w16: time from lower event (testing msb)
	btfss	STATUS,Z		;if preceding time >255:
	bra	SearchSlot_2		;lower margin ok, just upper margin might cause problem

	movlw	TMARGIN			;lower time margin too small?
	subwf	w16,W
	btfsc	STATUS,C
	bra	SearchSlot_2		;lower margin ok, just upper margin might cause problem
					;(W = [-TMARGIN ... -1])
	subwf	displace,F		;step forward by the difference in order to
	btfsc	STATUS,C		;increase lower margin to match TMARGIN
	incf	displace+1,F		;-> lower margin is ok now!

	addlw	-TMARGIN		;+ min. margin to next event 
	bra	SearchSlot_2a		;W = -(TMARGIN + displacement(lower margin))

SearchSlot_2
	movlw	-TMARGIN		;(no displacement from lower margin)
					
SearchSlot_2a
	movf	delta+1,F		;delta: time till next upper event (testing msb)
	btfss	STATUS,Z
	return				;upper time >255: both margins are big enough

	addwf	delta,W			;(W is negative) subtract from time till next event
	btfsc	STATUS,C
	return				;result >= 0:displace+TMARGIN does not collide with next event

;(W = -time by which displace+TMARGIN exceeds next event)
	addlw	TMARGIN*2		;add another 2*TMARGIN to get behind next event (incl. proper margin):		
	addwf	displace,F		;displace = T(next event) + TMARGIN	
	btfsc 	STATUS,C
	incf	displace+1,F

	clrf	w16+1
	movlw	TMARGIN			
	movwf	w16			
	call	queuePrepareInsert	;new insert-pos.= event(FSR1) + TMARGIN
	btfss	STATUS,C
	bra	SearchSlot		;still not sufficient margin here!
	return


;==================== Subroutine: remove event (W) from PWM queue 
;(W = channel*2 for leading edge, W = channel*2+1 for trailing edge)
;saves dtime from previous event (FSR0) in w16
PqUnlink
	call	queueFSR
	movf	FSR0L,W			
	btfsc	STATUS,Z
	return				;element is already unlinked!
	
	moviw	q_timeL[FSR0]		
	movwf	temp			;= dtime from previous event
	movwf	w16
	moviw	q_timeH[FSR0]
	movwf	temp+1
	movwf	w16+1
	moviw	q_timeL[FSR1]		
	addwf	temp,F			;+ dtime from current event
	moviw	q_timeH[FSR1]
	addwfc	temp+1,F		;= new dtime for previous queue element

	bcf	INTCON,TMR0IE		;no PWM interrupt during manipulation of queue!
	
	movf	pwm_queue_nxt,W		
	xorwf	FSR1L,W			;next event == unlinked element?
	btfss	STATUS,Z
	bra	PqUnlink_x		;no, so we can simply remove it from queue

	moviw	q_next[FSR1]		
	movwf	pwm_queue_nxt		;adjust pwm_queue_nxt: skip unlinked event

	bcf	INTCON,GIE		;no delays for next instructions!
	incfsz	TMR0,W			;if TMR1 == 0xFF, after (subwf TMR0,F) TMR0IF and carry may be set
	bra	$+2			;from overflow right between read-write (found out by testing)
	bra	$-2			;avoid race condition if timer is about to overflow

	moviw	q_timeL[FSR1]	
	subwf	TMR0,F			;adjust timer: add dtime of unlinked element 
	btfsc	INTCON,TMR0IF
	incf	tmr0h,F			;tmr0 did overflow!
	bcf	INTCON,TMR0IF
	bsf	INTCON,GIE		;(end of critical section)
	moviw	q_timeH[FSR1]	
	subwfb	tmr0h,F	
	
PqUnlink_x
	movf	INDF1,W			;next element -> W
	clrf	INDF1			;unlink this element
	movwf	INDF0			;change previous element's link
	movf	temp,W
	movwi	q_timeL[FSR0]		;adjust dtime of previous element
	movf	temp+1,W
	movwi	q_timeH[FSR0]	

	bsf	INTCON,TMR0IE		;re-enable PWM interrupt processing
;	bsf	INTCON,GIE		;re-enable PWM interrupt processing
	return

;====================== Displace leading edge event
;input: 
;	displace = displacement time

DisplaceLE
	aslf	channel,W
	call	PqUnlink		;FSR0=previous ev., w16=dtime (prev.<->removed)
	movf	FSR0L,W			;previous event is new origin
	movwf	FSR1L
	movf	displace,W
	addwf	w16,F			;add displacement time
	movf	displace+1,W
	addwfc	w16+1,F
	call	queuePrepareInsert	;get new location data (FSR0,w16,x)
	call	queueChFSR1		
;(fall through into PqLink)

;====================== Insert new event into queue
;input: 
;	FSR0 = previous event (set by queuePrepareInsert)
;	FSR1 = event to be linked into queue
;	w16 = dtime from last event to this event (by queuePrepareInsert)
;	delta = dtime to next event (by queuePrepareInsert)
PqLink
	movf	delta,W
	movwi	q_timeL[FSR1]		;store dtime (new event -> next event
	movf	delta+1,W
	movwi	q_timeH[FSR1]	
	moviw	q_next[FSR0]		;get old link from previous event
	movwi	q_next[FSR1]		;link this to next event
	bcf	INTCON,TMR0IE		;disable int during manipulation of active queue
	subwf	pwm_queue_nxt,W		;is this event the next one? 
	btfss	STATUS,Z
	bra	PqjustLink		;no

	bcf	INTCON,GIE		;no interrupts in next instructions
	incfsz	TMR0,W
	bra	$+2	
	bra	$-2			;avoid race condition if timer is about to overflow

	movf	TMR0,W			;remaining time until next event > delta?
	addwf	delta,F
	btfsc	INTCON,TMR0IF
	incf	delta+1,F		;TMR0 overflow!
	bsf	INTCON,GIE		;(end of critical section)

	movf	tmr0h,W
	addwfc	delta+1,F
	btfsc	STATUS,C
	bra	PqjustLink		;remaining time < delta

	clrf	TMR0
	bcf	INTCON,TMR0IF
	movf	delta,W
	movwf	TMR0			;set timer = time till processing of linked event
	movf	delta+1,W
	movwf	tmr0h
	movf	FSR1L,W
	movwf	pwm_queue_nxt		;this is next event!

PqjustLink
	movf	w16,W
	movwi	q_timeL[FSR0]		;set new dtime (previous event -> new event)
	movf	w16+1,W
	movwi	q_timeH[FSR0]
	movf	FSR1L,W
	movwi	q_next[FSR0]		;link previous event to new event

	bsf	INTCON,TMR0IE
	return


	end