;**************************************************************
;***      Universal Light Controller:                       ***
;***      Routines for evaluation of RC pulses              ***
;**************************************************************

#include picalic_config.inc

 if RCCHAN != 0
#include MCU_Headers_em.inc	;Pocessor specific definitions
#include macros.inc

	global rc_isr,procRCIn,rcdetect,rcval, rc_init, memswtch
	extern temp,w16,mscnt

RCIN_DATA udata
rc_ms100 	res 1		;used to generate 100ms intervals
errcnt 		res 1		;counts invalid rc-pulses
				;lower nibble: RX0, higher nibble: RX1
rctimer 	res 1		;1/10 sec. timer for RC 
				;lower nibble: RX0, higher nibble: RX1
rcdetect 	res 1		;MSB(nibble) = 1: no valid RC signal detected
rcval 		res RCCHAN	;normalized value (64 for 1.5 ms pulse width)
rclast 		res RCCHAN	;last pulse value

memswtch 	res 1		;bit 0..3 = RC memory switch bits, toggled by:
				;bit 0: stick moved down for short time
				;bit 1: stick moved down for long time
				;bit 2: stick moved up for short time
				;bit 3: stick moved up for long time
				;(higher nibble: same for RX1

t_rising0 	res 2		;timestamp: rising edge
rx0PW 		res 2		;latest pulse width (RX0)
 if RCCHAN > 1
t_rising1 	res 2		;data for second channel
rx1PW 		res 2		;latest pulse width (RX1)
 endif

	udata_shr
rc_flags res 1

;--------- rc_flags bit definitions:
#define RX0_EDGE 0		;rising or falling edge selection
#define RX1_EDGE 4
#define RX0_NEWP 1		;flag: got a new pulse 
#define RX1_NEWP 5
#define RX0_INIT 2		;initialization: wait for stick in middle pos.
#define RX1_INIT 6		;before processing memory switch 

 errorlevel-302
 
PROG0	code
;**************************************************************
;RC Interrupt Service Routine
;**************************************************************
;RC input level changed: store dedicated timestamp value 
;(bank0 is already selected)
rc_isr
	movf	TMR1H,W
	movwf	FSR1H			;use FSR1 for temporary storage
	movf	TMR1L,W
	movwf	FSR1L
	movf	TMR1H,W
	xorwf	FSR1H,W
	btfss	STATUS,Z
	bra	rc_isr			;TMR1H changed due to TMR1L overflow: read again!

	banksel	IOCFLG
	btfss	IOCFLG,PIN_RX0		;IOC @RX0-Pin?
	bra	rx_isrRx1		;no, maybe RX1

	btfss	rc_flags,RX0_EDGE	;expected rising edge?
	bra	rcisrCh0Falling		;no, should be fallig edge	

rcisrCh0Rising
	banksel	PORT_RX0
	btfss	PORT_RX0,PIN_RX0
	bra	rcisrCh0ClrIF		;port is low - is was a glitch

	bcf	rc_flags,RX0_EDGE	;next: wait for falling edge of pulse

	banksel	t_rising0
	movf	FSR1L,W
	movwf	t_rising0		;= time stamp (rising edge)
	movf	FSR1H,W
	movwf	t_rising0+1
	bra	rcisrCh0ClrIF

rcisrCh0Falling
	banksel	PORT_RX0
	btfsc	PORT_RX0,PIN_RX0
	bra	rcisrCh0ClrIF		;port is high - is was a glitch

	bsf	rc_flags,RX0_EDGE	;next: wait for rising edge of pulse
	btfsc	rc_flags,RX0_NEWP
	bra	rcisrCh0ClrIF		;previous PW was not yet read by main

	banksel	t_rising0
	movf	t_rising0,W
	subwf	FSR1L,W			;calculate PW = current time - t_rising
	movwf	rx0PW
	movf	t_rising0+1,W
	subwfb	FSR1H,W
	movwf	rx0PW+1
	bsf	rc_flags,RX0_NEWP	;new PW available!

rcisrCh0ClrIF
	banksel	IOCFLG
	bcf	IOCFLG,PIN_RX0
;eventually IOC on Rx1-Pin, too:

rx_isrRx1
 if RCCHAN > 1
	btfss	IOCFLG,PIN_RX1		;IOC @RX1-Pin?
	retfie				;no

	btfss	rc_flags,RX1_EDGE	;expected rising edge?
	bra	rcisrCh1Falling		;no, should be fallig edge	
	
rcisrCh1Rising
	banksel	PORT_RX1
	btfss	PORT_RX1,PIN_RX1
	bra	rcisrCh1ClrIF		;port is low - is was a glitch

	bcf	rc_flags,RX1_EDGE	;next: wait for falling edge of pulse

	banksel	t_rising1
	movf	FSR1L,W
	movwf	t_rising1		;= time stamp (rising edge)
	movf	FSR1H,W
	movwf	t_rising1+1
	bra	rcisrCh1ClrIF

rcisrCh1Falling
	banksel	PORT_RX1
	btfsc	PORT_RX1,PIN_RX1
	bra	rcisrCh1ClrIF		;port is high - is was a glitch

	bsf	rc_flags,RX1_EDGE	;next: wait for rising edge of pulse
	btfsc	rc_flags,RX1_NEWP
	bra	rcisrCh1ClrIF		;previous PW was not yet read by main

	banksel	t_rising1
	movf	t_rising1,W
	subwf	FSR1L,W			;calculate PW = current time - t_rising
	movwf	rx1PW
	movf	t_rising1+1,W
	subwfb	FSR1H,W
	movwf	rx1PW+1
	bsf	rc_flags,RX1_NEWP	;new PW available!

rcisrCh1ClrIF
	banksel	IOCFLG
	bcf	IOCFLG,PIN_RX1
 endif
	retfie

;========= Initialization of RC input(s)
;finds out, which RC inputs receive rc pulses
rc_init
	clrf	rc_flags
	banksel	memswtch
	movlw	0xFF
	movwf	errcnt
	movwf	rctimer
	movlw	1<<RX0_INIT | 1<<RX1_INIT
	movwf	rc_flags
	movlw	STARTSW
	movwf	memswtch
   ifdef RCV1INIT
	movlw	RCV1INIT
	movwf	rcval		;set to initial value
   endif
   if RCCHAN>1
     ifdef RCV2INIT
	movlw	RCV2INIT
	movwf	rcval+1		;set to initial value
     endif
	movlw	1<<PIN_RX0 | 1<<PIN_RX1
   else
	movlw	1<<PIN_RX0
   endif
	banksel	IOCFLG
	movwf	IOCP
	movwf	IOCN
	clrf	IOCFLG
 
	return

;========================== compute RC-values from received pulses
;called regularly 
#define RCADJ 22			;value to adjust for middle = 64

procRCIn
	banksel rc_ms100
	movf	mscnt,W
	subwf	rc_ms100,W		;100 ms over?
	btfss	WREG,7			;if MSB=1: not yet
	call	rcService10		;perform 100ms tasks

	btfss	rc_flags,RX0_NEWP	;got a pulse here?
	goto	procRCIn2
	lsrf	rx0PW+1,W		;y: process RX0 input pulse
	movwf	w16+1			;PW/2 -> w16
	rrf	rx0PW,W
	movwf	w16
	call	procRxP			;process PW value (RX0)
	bcf	rc_flags,RX0_NEWP	;processing done, ready for next pulse
procRCIn2
 if RCCHAN >1				;two rc-inputs:
	btfss	rc_flags,RX1_NEWP	;got a pulse here?
	return
	lsrf	rx1PW+1,W		;y: process RX1 input pulse
	movwf	w16+1			;PW/2 -> w16 (resolution: 2s)
	rrf	rx1PW,W
	movwf	w16

	call	rcswap			;swap RX1<->RX0 data
	call	procRxP			;process PW value (RX1)
	call	rcswap			;swap back
	bcf	rc_flags,RX1_NEWP	;processing done, ready for next pulse
 endif
	return


procRxP
	movf	rcval,W		;current value for processing of memory switch
	call	rcswpos		;get current switch-status (-1/0/+1)
	movwf	temp		;save here

	movlw	low 400		;subtract min. pulse with = 0.8 ms from PW
	subwf	w16,F
	movlw	high 400
	subwfb	w16+1,F
	btfss	STATUS,C	;negative result?
	goto	outrange	;y: pulse is too short

	lsrf	w16+1,F
	rrf	w16,F		;PW/4: resolution = 4s
	lsrf	w16+1,W
	rrf	w16,F		;PW/8: resolution = 8s
	andlw	0xFF
	btfss	STATUS,Z
	goto	outrange	;pulse is too long
	movf	w16,W		;pw-0.8ms -> W (8s units)
	movwf	temp+1		;1..2 ms = 25..150

	subwf	rclast,W	;compare previous with current value for error det.
	btfsc	WREG,7	
	sublw	0		;neg.-> invert
	andlw	0xF0		;Z=0 if large difference to previous pulse
	btfsc	STATUS,Z
	goto	rcvalok		;value ok.

	bsf	errcnt,0	;pulse very different from last pulse,
	goto	rcskip		;maybe it's an error

;Current pulse seems to be ok. 
;If there was an error before (errcnt != 0): need some consecutive right pulses 
rcvalok
	movlw	0xF0		;reset detection timer = 0.5s
	andwf	rcdetect,F	;(keep upper nibble unaffected)
	bsf	rcdetect,2	;lower nibble = 4
	
	movlw	0x0F
	andwf	errcnt,W
	btfsc	STATUS,Z
	goto	rcnewval	;errcnt == 0: assume pulse is valid

	decf	errcnt,F	;just wait for more valid pulses
	goto	rcskip

;received pulse width is in valid range!	
rcnewval
	movf	rcval,W		;get difference from last valid RC-value...
	addlw	RCADJ		;=> "raw" value of current rcval
	subwf	temp+1,W	;new-old val
	btfsc	STATUS,C
	goto	rcnvge		;new val >= old val 
;new val < old val:
	addlw	RCHYST
	btfsc	STATUS,C
	goto	rcskip		;difference < hysteresis: keep old value

	movlw	RCADJ-RCHYST
	goto	rcnewval1	;new val = recvd.val + hysteresis

;new value >= old value:
rcnvge
	addlw	-RCHYST
	btfss	STATUS,C
	goto	rcskip		;difference < hysteresis: keep old value

	movlw	RCADJ+RCHYST	;new val = recvd.val - hysteresis
rcnewval1
	subwf	temp+1,W	;adj. for middle pos. = 64 (+/- hysteresis)
	btfss	STATUS,C
	movlw	0		;<0 -> set to 0
	movwf	rcval		;store result -> new rcval

	call	rcswpos		;get new status of switch

	btfss	rc_flags,RX0_INIT
	goto	rcsw_ninit	;normal op.

	andlw	0xFF		;Z=1 for switch in middle pos
	btfss	STATUS,Z
	goto	rcskip		;do nothing until stick was in middle pos.
	
	bcf	rc_flags,RX0_INIT	

rcsw_ninit
	subwf	temp,W		
	btfss	STATUS,Z	;any change from last status?
	call	domemsw		;y: process memory switch 

rcskip
	movf	temp+1,W
	movwf	rclast
	return

outrange
	bsf	errcnt,1
	return

;========= status of rc switch changed: process memory switch
;temp = previous status of switch (-1/0/1) 
domemsw
	movf	temp,W		
	btfsc	STATUS,Z	;old pos = 0 (middle)?
	goto	doms_act	;y: change 0->1 or 0-> -1	

;switch was released from active position:
	movf	rctimer,W
	andlw	0x0F		;mask bits of this channel
	addlw	-MS_LONG
	btfsc	STATUS,C
	goto	doms_act	;rel. after long time, was already processed!

;switch released after short time: toggle primary ms-bit
	movlw	B'0001'		;if sw was -1: toggle bit 0
	btfss	temp,7	
	movlw	B'0010'		;sw was +1: toggle bit 1		 
	xorwf	memswtch,F 

doms_act
	movf	rcval,W
	call	rcswpos		;W= -1/0/1 (new position of switch)
	andlw	0xFF
	btfsc	STATUS,Z
	return			;new pos.= middle: done!

	movlw	0xF0
	andwf	rctimer,F	;start timer
	return
	

;========= service routine, called every 1/10th sec.
rcService10
	movlw	100
	addwf	rc_ms100,F	;reload 100ms timer

 if RCCHAN >1
	call	rcserv10a	;process ch 2, then swap back and process ch1
rcserv10a
	call	rcswap		;data -> lower nibbles

	decf	rcdetect,W	;--rcdetect -> W
	xorwf	rcdetect,W	;changed bits = 1
	andlw	0x0F		;affect only lower nibble
	btfss	rcdetect,3	;skip if it was already decremented to 0xXF
	xorwf	rcdetect,F	;adjust to decremented value
 else
	btfss	rcdetect,3	;decrement if >= 0
	decf	rcdetect,F	;runs to -1 if no RC pulses are detected
 endif

	movf	rctimer,W
	andlw	0x0F
	addlw	-MS_LONG
	btfsc	STATUS,C
	return			;time was already over (== MS_LONG), nothing more to do here!
	incf	rctimer,F
	addlw	1		
	btfss	STATUS,Z	;reached time MS_LONG after increment?
	return			;not yet

;timeout (MS_LONG):
	movf	rcval,W		;rcval->W 
	call	rcswpos		;-1/0/+1
	andlw	0xFF
	btfsc	STATUS,Z
	return			;== 0: switch in middle pos.-> exit

	addlw	1		;C=1 if switch pos = -1
	movlw	B'0100'		;if sw = -1: toggle bit 2
	btfss	STATUS,C
	movlw	B'1000'		;sw = +1: toggle bit 3
	xorwf	memswtch,F
	return

 if RCCHAN >1
;========== Swaps RC parameters (RX0<->RX1)
rcswap
	swapf	errcnt,F	
	swapf	rctimer,F
	swapf	memswtch,F
	swapf	rcdetect,F
	SWAPFF	rclast,rclast+1	;exchange rclast, rclast+1	
	SWAPFF	rcval,rcval+1	;same for rcval	
	return
 endif

;========= returns switch-position (-1/0/1) from value (W)
rcswpos
	addlw	-RCPW_LOW
	btfss	STATUS,C
	retlw 	-1		;<RCPW_LOW
	addlw	-(RCPW_HIGH-RCPW_LOW)
	btfss	STATUS,C
	retlw 	0		;<RCPW_HIGH
	retlw	1		;>= RC_PWHIGH



 endif		;RCCHAN != 0
	END
